import numpy as np
from colorsys import hsv_to_rgb
from collections import deque

from .util      import is_sequence, is_shape, tracked_array, Cache
from .constants import log

COLORS = {'red'    : [205,59,34],
          'purple' : [150,111,214],
          'blue'   : [119,158,203],
          'brown'  : [160,85,45]}
COLOR_DTYPE = np.dtype(np.uint8)
DEFAULT_COLOR = COLORS['purple']

class VisualAttributes(object):
    '''
    Hold the visual attributes (usually colors) for a mesh. 
    '''
    def __init__(self, mesh=None, **kwargs):
        self.mesh = mesh
        self._set = {'face'   : False,
                     'vertex' : False}
                     
        colors = _kwargs_to_color(mesh, **kwargs)
        self.vertex_colors, self.face_colors = colors

        # cache computed values which are cleared when
        # self.md5() changes, forcing a recompute
        self._cache = Cache(id_function = self.md5)    

    @property
    def defined(self):
        defined = np.any(self._set.values()) and self.mesh is not None
        return defined

    @property
    def transparency(self):
        cached = self._cache.get('transparency')
        if cached is not None: 
            return cached
        transparency = False
        color_max = (2**(COLOR_DTYPE.itemsize*8)) - 1
        if self._set['face']:
            transparency = (self._face_colors.ndim == 2 and
                            self._face_colors.shape[1] == 4 and
                            np.any(self._face_colors[:,3] < color_max))
        elif self._set['vertex']:
            transparency = (self._vertex_colors.ndim == 2 and
                            self._vertex_colors.shape[1] == 4 and
                            np.any(self._vertex_colors[:,3] < color_max))

        return self._cache.set(key   = 'transparency',
                               value = bool(transparency))

    def md5(self):
        md5 = ''
        if hasattr(self.mesh, 'md5'):
            md5 += self.mesh.md5()            
        if self._set['face']:
            md5 += self._face_colors.md5()
        if self._set['vertex']:
            md5 += self._vertex_colors.md5()
        return md5

    @property
    def face_colors(self):
        if not (is_sequence(self._face_colors) and
                len(self._face_colors) == len(self.mesh.faces)):
            self._face_colors = tracked_array(np.tile(DEFAULT_COLOR,
                                                      (len(self.mesh.faces), 1)))
            log.debug('Setting default colors for faces')
        return self._face_colors
        
    @face_colors.setter
    def face_colors(self, values):
        values = np.asanyarray(values)
        if values.shape in ((3,), (4,)):
            # case where a single RGB/RGBa color has been passed to the setter
            # we apply this color to all faces 
            values = np.tile(values, (len(self.mesh.faces), 1))

        self._face_colors = tracked_array(values)
        # this will only consider colors defined if they were passed as a sequence
        # is_sequence gracefully handles both None and np.array(None)
        self._set['face'] = is_sequence(values)

    @property
    def vertex_colors(self):
        if not (is_sequence(self._vertex_colors) and
                len(self._vertex_colors) == len(self.mesh.vertices)):
            log.debug('Vertex colors being generated from face colors')
            self._vertex_colors = face_to_vertex_color(self.mesh, 
                                                       self.face_colors)
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, values):
        self._vertex_colors = tracked_array(values)
        self._set['vertex'] = is_sequence(values)

    def update_faces(self, mask):
        if not self._set['face']: 
            return
        try: self._face_colors = self._face_colors[mask]
        except: log.warning('Face colors not updated', exc_info=True) 
            
    def update_vertices(self, mask):
        if not self._set['vertex']:
            return
        try:    
            self.vertex_colors = self._vertex_colors[mask]
        except: 
            log.debug('Vertex colors not updated', exc_info=True)

    def subsets(self, faces_sequence):
        result = deque()
        for f in faces_sequence:
            result.append(VisualAttributes())
            if self._set['face']:
                result[-1].face_colors = self.face_colors[np.array(list(f))]
        return np.array(result)

    def union(self, others):
        return visuals_union(np.append(self, others))

def _kwargs_to_color(mesh, **kwargs):
    '''
    Given a set of keyword arguments, see if any reference color
    in their name, and match the dimensions of the mesh.
    '''

    def pick_option(vf):
        if any(i is None for i in vf):
            return vf
        result = [None, None]
        signal = [i.ptp(axis=0).sum() for i in vf]
        signal_max = np.argmax(signal)
        result[signal_max] = vf[signal_max]
        return result
    def pick_color(sequence):
        if len(sequence) == 0:
            return None
        elif len(sequence) == 1:
            return sequence[0]
        else:
            signal = [i.ptp(axis=0).sum() for i in sequence]
            signal_max = np.argmax(signal)
            return sequence[signal_max]

    if mesh is None:
        result = [None, None]
        if 'face_colors' in kwargs:
            result[1] = np.asanyarray(kwargs['face_colors'])
        if 'vertex_colors' in kwargs:
            result[0] = np.asanyarray(kwargs['vertex_colors'])
        return result
        
    vertex = deque()
    face   = deque()
    
    for key in kwargs.keys():
        if not ('color' in key): 
            continue
        value = np.asanyarray(kwargs[key])
        if len(value) == len(mesh.vertices):
            vertex.append(value)
        elif len(value) == len(mesh.faces):
            face.append(value)
    return pick_option([pick_color(i) for i in [vertex, face]])

def visuals_union(visuals, *args):
    visuals = np.append(visuals, args)
    color = {'face_colors'   : None,
             'vertex_colors' : None}
    if all(is_shape(i._face_colors, (-1,(3,4))) for i in visuals):
        color['face_colors'] = np.vstack([_rgba(i._face_colors) for i in visuals])
    if all(is_shape(i._vertex_colors, (-1,(3,4))) for i in visuals):
        color['vertex_colors'] = np.vstack([i._vertex_colors for i in visuals])
    return VisualAttributes(**color)

def _rgba(colors, dtype=COLOR_DTYPE):
    colors = np.asanyarray(colors)
    if is_shape(colors, (-1,3)):
        opaque = (2**(np.dtype(dtype).itemsize * 8)) - 1
        colors = np.column_stack((colors,
                                  opaque * np.ones(len(colors))))
    return colors.astype(dtype)

def random_color(dtype=COLOR_DTYPE):
    '''
    Return a random RGB color using datatype specified.
    '''
    hue  = np.random.random() + .61803
    hue %= 1.0
    color = np.array(hsv_to_rgb(hue, .99, .99))
    if np.dtype(dtype).kind in 'iu':
        max_value = (2**(np.dtype(dtype).itemsize * 8)) - 1
        color    *= max_value
    color     = color.astype(dtype)
    return color

def face_to_vertex_color(mesh, face_colors, dtype=COLOR_DTYPE):
    '''
    Convert a set of face colors into a set of vertex colors.
    '''
    
    color_dim = np.shape(face_colors)[1]

    vertex_colors = np.zeros((len(mesh.vertices),3,color_dim))
    population    = np.zeros((len(mesh.vertices),3), dtype=np.bool)

    vertex_colors[[mesh.faces[:,0],0]] = face_colors
    vertex_colors[[mesh.faces[:,1],1]] = face_colors
    vertex_colors[[mesh.faces[:,2],2]] = face_colors

    population[[mesh.faces[:,0], 0]] = True
    population[[mesh.faces[:,1], 1]] = True
    population[[mesh.faces[:,2], 2]] = True

    # clip the population sum to 1, to avoid a division error in edge cases
    populated     = np.clip(population.sum(axis=1), 1, 3)
    vertex_colors = vertex_colors.sum(axis=1) / populated.reshape((-1,1))
    return vertex_colors.astype(dtype)



import numpy as np
import networkx as nx

from collections import deque

from .constants import log, tol, MeshError
from .grouping  import group, group_rows, boolean_rows
from .geometry  import faces_to_edges
from .points    import unitize
from .util      import diagonal_dot, is_sequence

try: from scipy.spatial import cKDTree as KDTree
except ImportError: log.warning('Scipy unavailable')

try: 
    from graph_tool import Graph as GTGraph
    from graph_tool.topology import label_components
    _has_gt = True
except: 
    _has_gt = False
    log.warning('No graph-tool! Some operations will be much slower!')

def face_adjacency(faces, return_edges=False):
    '''
    Returns an (n,2) list of face indices.
    Each pair of faces in the list shares an edge, making them adjacent.

    This is useful for lots of things, for example finding connected subgraphs:

    graph = nx.Graph()
    graph.add_edges_from(mesh.face_adjacency)
    groups = nx.connected_components(graph_connected.subgraph(interesting_faces))
    '''

    # first generate the list of edges for the current faces
    # also return the index for which face the edge is from
    edges, edge_face_index = faces_to_edges(faces, return_index = True)
    edges.sort(axis=1)
    # this will return the indices for duplicate edges
    # every edge appears twice in a well constructed mesh
    # so for every row in edge_idx, edges[edge_idx[*][0]] == edges[edge_idx[*][1]]
    # in this call to group rows, we discard edges which don't occur twice
    edge_groups = group_rows(edges, require_count=2)

    if len(edge_groups) == 0:
        log.error('No adjacent faces detected! Did you merge vertices?')

    # the pairs of all adjacent faces
    # so for every row in face_idx, self.faces[face_idx[*][0]] and 
    # self.faces[face_idx[*][1]] will share an edge
    face_adjacency = edge_face_index[edge_groups]
    if return_edges:
        face_adjacency_edges = edges[edge_groups[:,0]]
        return face_adjacency, face_adjacency_edges
    return face_adjacency

def adjacency_angle(mesh, angle):
    '''
    Return the adjacent faces of a mesh only if the faces
    are at less than a specified angle.

    Arguments
    ----------
    mesh: Trimesh object
    angle: angle in radians, faces at angles LARGER than this
           will be considered NOT adjacenct

    Returns
    ----------
    adjacency: (n,2) int list of face indices in mesh
    '''

    # use the cached adjacency if possible (n,2)
    adjacency = mesh.face_adjacency
    # normal vectors for adjacent faces (n, 2, 3)
    normals = mesh.face_normals[adjacency]
    # dot products of normals (n)
    dots = diagonal_dot(normals[:,0], normals[:,1])
    # clip for floating point error
    dots = np.clip(dots, -1.0, 1.0)
    adj_ok = np.abs(np.arccos(dots)) < angle
    # result is (m,2)
    result = adjacency[adj_ok]
    return result

def shared_edges(faces_a, faces_b):
    '''
    Given two sets of faces, find the edges which are in both sets.

    Arguments
    ---------
    faces_a: (n,3) int, set of faces
    faces_b: (m,3) int, set of faces

    Returns
    ---------
    shared: (p, 2) int, set of edges
    '''
    e_a = np.sort(faces_to_edges(faces_a), axis=1)
    e_b = np.sort(faces_to_edges(faces_b), axis=1)
    shared = boolean_rows(e_a, e_b, operation=set.intersection)
    return shared

def connected_edges(G, nodes):
    '''
    Given graph G and list of nodes, return the list of edges that 
    are connected to nodes
    '''
    nodes_in_G = deque()
    for node in nodes:
        if not G.has_node(node): continue
        nodes_in_G.extend(nx.node_connected_component(G, node))
    edges = G.subgraph(nodes_in_G).edges()
    return edges
 
def facets(mesh):
    '''
    Find the list of parallel adjacent faces.
    
    Arguments
    ---------
    mesh:  Trimesh
    
    Returns
    ---------
    facets: list of groups of face indexes (in mesh.faces) of parallel 
            adjacent faces. 
    '''
    def facets_nx():
        graph_parallel = nx.from_edgelist(face_idx[parallel])
        facets_idx = np.array([list(i) for i in nx.connected_components(graph_parallel)])
        return facets_idx
        
    def facets_gt():
        graph_parallel = GTGraph()
        graph_parallel.add_edge_list(face_idx[parallel])
        connected  = label_components(graph_parallel, directed=False)[0].a
        facets_idx = group(connected, min_len=2)
        return facets_idx

    # (n,2) list of adjacent face indices
    face_idx    = mesh.face_adjacency

    # test adjacent faces for angle
    normal_pairs = mesh.face_normals[[face_idx]]
    normal_dot   = (np.sum(normal_pairs[:,0,:] * normal_pairs[:,1,:], axis=1) - 1)**2

    # if normals are actually equal, they are parallel with a high degree of confidence
    parallel     = normal_dot < tol.zero
    non_parallel = np.logical_not(parallel)

    # saying that two faces are *not* parallel is susceptible to error
    # so we add a radius check which computes the distance between face
    # centroids and divides it by the dot product of the normals
    # this means that small angles between big faces will have a large
    # radius which we can filter out easily.
    # if you don't do this, floating point error on tiny faces can push
    # the normals past a pure angle threshold even though the actual 
    # deviation across the face is extremely small. 
    center      = mesh.triangles.mean(axis=1)
    center_sq = np.sum(np.diff(center[face_idx], 
                               axis = 1).reshape((-1,3)) ** 2, axis=1)
    radius_sq = center_sq[non_parallel] / normal_dot[non_parallel]
    parallel[non_parallel] = radius_sq > tol.facet_rsq

    # graph-tool is ~6x faster than networkx but is more difficult to install
    if _has_gt: return facets_gt()
    else:       return facets_nx()

def split(mesh, only_watertight=True, adjacency=None):
    '''
    Given a mesh, will split it up into a list of meshes based on face connectivity
    If only_watertight is true, it will only return meshes where each face has
    exactly 3 adjacent faces.

    Arguments
    ----------
    mesh: Trimesh 
    only_watertight: if True, only return watertight components
    adjacency: (n,2) list of face adjacency to override using the plain
               adjacency calculated automatically. 

    Returns
    ----------
    meshes: list of Trimesh objects
    '''

    def split_nx():
        adjacency_graph = nx.from_edgelist(adjacency)
        components = nx.connected_components(adjacency_graph)
        result = mesh.submesh(components, only_watertight=only_watertight)
        return result

    def split_gt():
        g = GTGraph()
        g.add_edge_list(adjacency)
        component_labels = label_components(g, directed=False)[0].a
        components = group(component_labels)
        result = mesh.submesh(components, only_watertight=only_watertight)
        return result

    if adjacency is None:
        adjacency = mesh.face_adjacency
    
    if _has_gt: 
        return split_gt()
    else:       
        return split_nx()

def smoothed(mesh, angle):
    '''
    Return a non- watertight version of the mesh which will
    render nicely with smooth shading. 

    Arguments
    ---------
    mesh: Trimesh object
    angle: float, radians 

    Returns
    ---------
    smooth: Trimesh object
    '''
    adjacency = adjacency_angle(mesh, angle)
    graph = nx.from_edgelist(adjacency)
    graph.add_nodes_from(np.arange(len(mesh.faces)))
    smooth = mesh.submesh(nx.connected_components(graph),
                          only_watertight = False,
                          append = True)
    return smooth

def is_watertight(edges):
    '''
    Arguments
    ---------
    edges: (n,2) int, set of vertex indices
    
    Returns
    ---------
    watertight: boolean, whether every edge is contained by two faces
    '''
    edges = np.sort(edges, axis=1)
    groups = group_rows(edges, require_count=2)
    watertight = (len(groups) * 2) == len(edges)
    return watertight



from .generic import MeshScript

from distutils.spawn import find_executable

import inspect
import os

def boolean(meshes, operation='difference'):
    operation = str.upper(operation)
    if operation == 'INTERSECTION': 
        operation = 'INTERSECT'

    current = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    script = open(os.path.join(current, '../templates/blender.py.template'),'rb').read()
    script = script.replace('$operation', operation)
  
    with MeshScript(meshes = meshes, 
                    script = script) as blend:
        result = blend.run('blender --background --python $script')

    # blender returns actively incorrect face normals
    result['face_normals'] = None
    return result

exists = find_executable('blender') is not None




'''
path.py

A library designed to work with vector paths.
'''

import numpy as np
import networkx as nx

from shapely.geometry import Polygon, Point
from scipy.spatial import cKDTree as KDTree
from copy import deepcopy
from collections import deque

from .simplify  import simplify_path
from .polygons  import polygons_enclosure_tree, is_ccw, medial_axis, polygon_hash, path_to_polygon, polygon_obb
from .traversal import vertex_graph, closed_paths, discretize_path
from .io.export import export_path

from ..points    import plane_fit, transform_points
from ..geometry  import plane_transform
from ..grouping  import unique_rows
from ..units     import _set_units
from ..util      import decimal_to_digits, is_sequence
from ..constants import log, time_function
from ..constants import tol_path as tol

from .. import util

class Path(object):
    '''
    A Path object consists of two things:
    vertices: (n,[2|3]) coordinates, stored in self.vertices
    entities: geometric primitives (lines, arcs, and circles)
              that reference indices in self.vertices
    '''
    def __init__(self, 
                 entities = [], 
                 vertices = [],
                 metadata = None):
        '''
        entities:
            Objects which contain things like keypoints, as 
            references to self.vertices
        vertices:
            (n, (2|3)) list of vertices
        '''
        self.entities = np.array(entities)
        self.vertices = vertices
        self.metadata = dict()

        if metadata.__class__.__name__ == 'dict':
            self.metadata.update(metadata)

        self._cache = util.Cache(id_function = self.md5)

        # literally nothing will work if vertices aren't merged properly
        self.merge_vertices()

    def process(self):
        log.debug('Processing drawing')
        with self._cache:     
            for func in self._process_functions():
                func()
        return self
        
    @property
    def vertices(self):
        return self._vertices
        
    @vertices.setter
    def vertices(self, values):
        self._vertices = util.tracked_array(values)

    def md5(self):
        result  = self.vertices.md5()
        result += str(len(self.entities))
        return result

    @property
    def paths(self):
        if 'paths' in self._cache:
            return self._cache.get('paths')
        with self._cache:
            paths = closed_paths(self.entities, self.vertices)
        return self._cache.set('paths', paths)

    @property
    def kdtree(self):
        if 'kdtree' in self._cache:
            return self._cache.get('kdtree')
        with self._cache:
            kdtree = KDTree(self.vertices.view(np.ndarray))
        return self._cache.set('kdtree', kdtree)

    @property
    def discrete(self):
        return self._cache.get('discrete')

    @property
    def scale(self):
        return np.max(np.ptp(self.vertices, axis=0))

    @property
    def bounds(self):
        return np.vstack((np.min(self.vertices, axis=0),
                          np.max(self.vertices, axis=0)))
    @property
    def extents(self):
        return np.diff(self.bounds, axis=0)[0]

    @property
    def units(self):
        if 'units' in self.metadata:
            return self.metadata['units']
        else:
            return None

    @property
    def is_closed(self):
        return all(i == 2 for i in self.vertex_graph.degree().values())

    @property
    def vertex_graph(self):
        if 'vertex_graph' in self._cache:
            return self._cache.get('vertex_graph')
        with self._cache:
            graph, closed = vertex_graph(self.entities)
        return self._cache.set('vertex_graph', graph)
   
    @units.setter
    def units(self, units):
        self.metadata['units'] = units
            
    def set_units(self, desired, guess=False):
        _set_units(self, desired, guess)

    def transform(self, transform):
        self.vertices = transform_points(self.vertices, transform)

    def rezero(self):
        self.vertices -= self.vertices.min(axis=0)
        
    def merge_vertices(self):
        '''
        Merges vertices which are identical and replaces references
        '''
        digits = decimal_to_digits(tol.merge * self.scale, min_digits=1)
        unique, inverse = unique_rows(self.vertices, digits=digits)
        self.vertices = self.vertices[unique]
        for entity in self.entities: 
            entity.points = inverse[entity.points]

    def replace_vertex_references(self, replacement_dict):
        for entity in self.entities: entity.rereference(replacement_dict)

    def remove_entities(self, entity_ids):
        '''
        Remove entities by their index.
        '''
        if len(entity_ids) == 0: return
        kept = np.setdiff1d(np.arange(len(self.entities)), entity_ids)
        self.entities = np.array(self.entities)[kept]

    def remove_invalid(self):
        valid = np.array([i.is_valid for i in self.entities], dtype=np.bool)
        self.entities = self.entities[valid]

    def remove_duplicate_entities(self):
        entity_hashes   = np.array([i.hash for i in self.entities])
        unique, inverse = unique_rows(entity_hashes)
        if len(unique) != len(self.entities):
            self.entities = np.array(self.entities)[unique]
            
    def referenced_vertices(self):
        referenced = deque()
        for entity in self.entities: 
            referenced.extend(entity.points)
        return np.array(referenced)
    
    def remove_unreferenced_vertices(self):
        '''
        Removes all vertices which aren't used by an entity
        Reindexes vertices from zero, and replaces references
        '''
        referenced       = self.referenced_vertices()
        unique_ref       = np.int_(np.unique(referenced))
        replacement_dict = dict()
        replacement_dict.update(np.column_stack((unique_ref, 
                                                 np.arange(len(unique_ref)))))
        self.replace_vertex_references(replacement_dict)
        self.vertices = self.vertices[[unique_ref]] 
        
    def discretize_path(self, path):
        '''
        Return a (n, dimension) list of vertices. 
        Samples arcs/curves to be line segments
        '''
        discrete = discretize_path(self.entities, 
                                   self.vertices, 
                                   path, 
                                   scale = self.scale)
        return discrete
        
    def export(self, file_type='dict', file_obj=None):
        return export_path(self, 
                           file_type = file_type,
                           file_obj  = file_obj)

    def to_dict(self):
        export_dict = self.export(file_type='dict')
        return export_dict

    def copy(self):
        return deepcopy(self)

    def show(self):
        if self.is_closed:
            self.plot_discrete(show=True)
        else:
            self.plot_entities(show=True)

    def __add__(self, other):
        new_entities = deepcopy(other.entities)
        for entity in new_entities:
            entity.points += len(self.vertices)
        new_entities = np.append(deepcopy(self.entities), new_entities)
 
        new_vertices = np.vstack((self.vertices, other.vertices))
        new_meta     = deepcopy(self.metadata)
        new_meta.update(other.metadata)

        new_path = self.__class__(entities = new_entities,
                                  vertices = new_vertices,
                                  metadata = new_meta)
        return new_path
   
class Path3D(Path):
    def _process_functions(self): 
        return [self.merge_vertices,
                self.remove_duplicate_entities,
                self.remove_unreferenced_vertices,
                self.generate_closed_paths,
                self.generate_discrete]
               
    def generate_discrete(self):
        discrete = list(map(self.discretize_path, self.paths))
        self._cache.set('discrete', discrete)

    def to_planar(self, to_2D=None, normal=None, check=True):
        '''
        Check to see if current vectors are all coplanar.
        
        If they are, return a Path2D and a transform which will 
        transform the 2D representation back into 3 dimensions
        '''
        if to_2D is None:
            C, N = plane_fit(self.vertices)
            if normal is not None:
                N *= np.sign(np.dot(N, normal))
            to_2D = plane_transform(C,N)
 
        flat = transform_points(self.vertices, to_2D)
        
        if check and np.any(np.std(flat[:,2]) > tol.planar):
            log.error('points have z with deviation %f', np.std(flat[:,2]))
            raise NameError('Points aren\'t planar!')
            
        vector = Path2D(entities = deepcopy(self.entities), 
                        vertices = flat[:,0:2])
        to_3D  = np.linalg.inv(to_2D)

        return vector, to_3D
 
    def plot_discrete(self, show=False):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        fig  = plt.figure()
        axis = fig.add_subplot(111, projection='3d')
        for discrete in self.discrete:
            axis.plot(*discrete.T)
        if show: plt.show()

    def plot_entities(self, show=False):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        fig  = plt.figure()
        axis = fig.add_subplot(111, projection='3d')
        for entity in self.entities:
            vertices = self.vertices[entity.points]
            axis.plot(*vertices.T)        
        if show: plt.show()

class Path2D(Path):
    def _process_functions(self): 
        return [self.merge_vertices,
                self.remove_duplicate_entities,
                self.remove_unreferenced_vertices]

    def apply_obb(self):
        if len(self.root) == 1:
            bounds, T = polygon_obb(self.polygons_closed[self.root[0]])
            self.transform(T)
        else:
            raise ValueError('Not implemented for multibody geometry')

    @property
    def body_count(self):
        return len(self.root)

    @property
    def polygons_full(self):
        if 'polygons_full' in self._cache:
            return self._cache.get('polygons_full') 

        with self._cache:
            result = [None] * len(self.root)
            for index, root in enumerate(self.root):
                hole_index = self.connected_paths(root, include_self=False)
                holes = [p.exterior.coords for p in self.polygons_closed[hole_index]]
                shell = self.polygons_closed[root].exterior.coords
                result[index] = Polygon(shell  = shell,
                                        holes  = holes)
        return self._cache.set('polygons_full', result)

    @property
    def area(self):
        '''
        Return the area of the polygons interior
        '''
        area = np.sum([i.area for i in self.polygons_full])
        return area
        
    def extrude(self, height, **kwargs):
        '''
        Extrude the current 2D path into a 3D mesh. 

        Arguments
        ----------
        height: float, how far to extrude the profile
        kwargs: passed directly to meshpy.triangle.build:
                triangle.build(mesh_info, 
                               verbose=False, 
                               refinement_func=None, 
                               attributes=False, 
                               volume_constraints=True, 
                               max_volume=None, 
                               allow_boundary_steiner=True, 
                               allow_volume_steiner=True, 
                               quality_meshing=True, 
                               generate_edges=None, 
                               generate_faces=False, 
                               min_angle=None)
        Returns
        --------
        mesh: trimesh object representing extruded polygon
        '''
        from ..creation import extrude_polygon
        result = [extrude_polygon(i, height, **kwargs) for i in self.polygons_full]
        if len(result) == 1: 
            return result[0]
        return result

    def medial_axis(self, resolution=None, clip=None):

        '''
        Find the approximate medial axis based
        on a voronoi diagram of evenly spaced points on the boundary of the polygon.

        Arguments
        ----------
        resolution: target distance between each sample on the polygon boundary
        clip:       [minimum number of samples, maximum number of samples]
                    specifying a very fine resolution can cause the sample count to
                    explode, so clip specifies a minimum and maximum number of samples
                    to use per boundary region. To not clip, this can be specified as:
                    [0, np.inf]

        Returns
        ----------
        medial:     Path2D object
        '''
        if 'medial' in self._cache:
            return self._cache.get('medial')

        if resolution is None:
            resolution = self.scale / 1000.0

        medials = [medial_axis(i, resolution, clip) for i in self.polygons_full]
        medials = np.sum(medials)
        return self._cache.set(key = 'medial',
                               value = medials)
        
    def connected_paths(self, path_id, include_self = False):
        if len(self.root) == 1:
            path_ids = np.arange(len(self.polygons_closed))
        else:
            path_ids = list(nx.node_connected_component(self.enclosure, path_id))
        if include_self: 
            return np.array(path_ids)
        return np.setdiff1d(path_ids, [path_id])
        
    def simplify(self):
        self._cache.clear()
        simplify_path(self)

    def split(self):
        '''
        If the current Path2D consists of n 'root' curves,
        split them into a list of n Path2D objects
        '''
        if self.root is None or len(self.root) == 0:
            split = []
        elif len(self.root) == 1:
            split = [deepcopy(self)]
        else:
            split = [None] * len(self.root)
            for i, root in enumerate(self.root):
                connected = self.connected_paths(root, include_self=True)
                new_root     = np.nonzero(connected == root)[0]
                new_entities = deque()
                new_paths    = deque()
                new_metadata = {'split_2D' : i}
                new_metadata.update(self.metadata)


                for path in self.paths[connected]:
                    new_paths.append(np.arange(len(path)) + len(new_entities))
                    new_entities.extend(path)
                new_entities = np.array(new_entities)
                # prevents the copying from nuking our cache
                with self._cache:
                    split[i] = Path2D(entities = deepcopy(self.entities[new_entities]),
                                      vertices = deepcopy(self.vertices))
                    split[i]._cache.update({'paths'          : np.array(new_paths),
                                            'polygons_closed': self.polygons_closed[connected],
                                            'root'           : new_root})
        [i._cache.id_set() for i in split]
        self._cache.id_set()
        return np.array(split)

    def plot_discrete(self, show=False, transform=None, axes=None):
        import matplotlib.pyplot as plt
        plt.axes().set_aspect('equal', 'datalim')
        def plot_transformed(vertices, color='g'):
            if transform is None: 
                if axes is None:
                    plt.plot(*vertices.T, color=color)
                else:
                    axes.plot(*vertices.T, color=color)
            else:
                transformed = transform_points(vertices, transform)
                plt.plot(*transformed.T, color=color)
        for i, polygon in enumerate(self.polygons_closed):
            color = ['g','k'][i in self.root]
            plot_transformed(np.column_stack(polygon.boundary.xy), color=color)
        if show: plt.show()

    def plot_entities(self, show=False):
        import matplotlib.pyplot as plt
        plt.axes().set_aspect('equal', 'datalim')
        eformat = {'Line0'  : {'color'  :'g', 'linewidth':1},
                   'Line1'  : {'color'  :'y', 'linewidth':1},
                   'Arc0'   : {'color'  :'r', 'linewidth':1}, 
                   'Arc1'   : {'color'  :'b', 'linewidth':1},
                   'Bezier0': {'color'  :'k', 'linewidth':1},
                   'BSpline0': {'color' :'m', 'linewidth':1},
                   'BSpline1': {'color' :'m', 'linewidth':1}}
        for entity in self.entities:
            discrete = entity.discrete(self.vertices)
            e_key    = entity.__class__.__name__ + str(int(entity.closed))
            plt.plot(discrete[:,0], 
                     discrete[:,1], 
                     **eformat[e_key])
        if show: plt.show()

    def identifier(self):
        if len(self.polygons_full) != 1: 
            raise TypeError('Identifier only valid for single body')
        return polygon_hash(self.polygons_full[0])


    @property
    def polygons_valid(self):
        exists = self.polygons_closed
        return self._cache.get('polygons_valid')

    @property
    def polygons_closed(self):
        if 'polygons_closed' in self._cache: 
            return self._cache.get('polygons_closed')

        def reverse_path(path):
            for entity in self.entities[path]: 
                entity.reverse()
            return path[::-1]
        with self._cache:
            polygons = [None]  * len(self.paths)
            valid    = [False] * len(self.paths)
            for i, path in enumerate(self.paths):
                discrete = discretize_path(self.entities, 
                                           self.vertices, 
                                           path, 
                                           scale = self.scale)
                candidate = path_to_polygon(discrete, scale=self.scale)
                if candidate is None: 
                    continue
                if not candidate.exterior.is_ccw:
                    log.debug('Clockwise polygon detected, correcting!')
                    self.paths[i] = reverse_path(path)
                    candidate = Polygon(np.array(candidate.exterior.coords)[::-1])
                polygons[i] = candidate
                valid[i]    = True
            valid = np.array(valid, dtype=np.bool)
            polygons = np.array(polygons)[valid]

        self._cache.set('polygons_valid',  valid)
        self._cache.set('polygons_closed', polygons)
        return polygons

    @property
    def root(self):
        if 'root' in self._cache:
            return self._cache.get('root')
        with self._cache:
            root, enclosure = polygons_enclosure_tree(self.polygons_closed)
        self._cache.set('enclosure_directed', enclosure)
        return self._cache.set('root', root)

    @property
    def enclosure(self):
        if 'enclosure' in self._cache:
            return self._cache.get('enclosure')
        with self._cache:
            undirected = self.enclosure_directed.to_undirected()
        return self._cache.set('enclosure', undirected)
        
    @property
    def enclosure_directed(self):
        if 'enclosure_directed' in self._cache:
            return self._cache.get('enclosure_directed')
        with self._cache:
            root, enclosure = polygons_enclosure_tree(self.polygons_closed)
        self._cache.set('root', root)
        return self._cache.set('enclosure_directed', enclosure)



import numpy as np
from collections import deque

from networkx      import from_edgelist, connected_components

from .points    import unitize
from .util      import decimal_to_digits, vector_to_spherical, spherical_to_vector
from .constants import log, tol

try: from scipy.spatial import cKDTree as KDTree
except ImportError: log.warning('Scipy unavailable')
    
def merge_vertices_hash(mesh):
    '''
    Removes duplicate vertices, based on integer hashes.
    This is roughly 20x faster than querying a KD tree in a loop
    '''
    pre_merge = len(mesh.vertices)
    unique, inverse = unique_rows(mesh.vertices)
    mesh.update_vertices(unique, inverse)

def merge_vertices_kdtree(mesh, angle=None):
    '''
    Merges vertices which are identical, AKA within 
    Cartesian distance TOL_MERGE of each other.  
    Then replaces references in mesh.faces
    
    If max_angle == None, vertex normals won't be looked at. 
    if max_angle has a value, vertices will only be considered identical
    if they are within TOL_MERGE of each other, and the angle between
    their normals is less than angle_max

    Performance note:
    cKDTree requires scipy >= .12 for this query type and you 
    probably don't want to use plain python KDTree as it is crazy slow (~1000x in tests)
    '''

    tree = mesh.kdtree()
    used    = np.zeros(len(mesh.vertices), dtype=np.bool)
    inverse = np.arange(len(mesh.vertices), dtype=np.int)
    unique  = deque()

    vectors = mesh.vertex_normals

    for index, vertex in enumerate(mesh.vertices):
        if used[index]: 
            continue

        neighbors = np.array(tree.query_ball_point(mesh.vertices[index], 
                                                   tol.merge))
        if angle is not None:
            groups = group_vectors(vectors[neighbors], angle=angle)[1]
        else:
            groups = np.arange(len(neighbors)).astype(int).reshape((1,-1))
                            
        used[neighbors] = True
        for group in groups:
            inverse[neighbors[group]] = len(unique)
            unique.append(neighbors[group[0]])
    mesh.update_vertices(np.array(unique), inverse)
   
def replace_references(data, reference_dict):
    '''
    Replace elements in an array as per a dictionary of replacement values. 

    Arguments
    ----------
    data:           numpy array 
    reference_dict: dictionary of replacement value mapping, eg: {2:1, 3:1, 4:5}
    '''
    shape = np.shape(data)
    view  = np.array(data).view().reshape((-1))
    for i, value in enumerate(view):
        if value in reference_dict:
            view[i] = reference_dict[value]
    return view.reshape(shape)

def group(values, min_len=0, max_len=np.inf):
    '''
    Return the indices of values that are identical
    
    Arguments
    ----------
    values:     1D array 
    min_len:    int, the shortest group allowed
                All groups will have len >= min_length
    max_len:    int, the longest group allowed
                All groups will have len <= max_length
    
    Returns
    ----------
    groups: sequence of indices to form groups
            IE [0,1,0,1] returns [[0,2], [1,3]]
    '''
    order    = values.argsort()
    values   = values[order]
    dupe     = np.greater(np.abs(np.diff(values)), tol.zero)
    dupe_idx = np.append(0, np.nonzero(dupe)[0] + 1)
    dupe_len = np.diff(np.hstack((dupe_idx, len(values)))) 
    dupe_ok  = np.logical_and(np.greater_equal(dupe_len, min_len),
                              np.less_equal(   dupe_len, max_len))
    groups   = [order[i:(i+j)] for i, j in zip(dupe_idx[dupe_ok], dupe_len[dupe_ok])]
    return groups
    
def hashable_rows(data, digits=None):
    '''
    We turn our array into integers, based on the precision 
    given by digits, and then put them in a hashable format. 
    
    Arguments
    ---------
    data:    (n,m) input array
    digits:  how many digits to add to hash, if data is floating point
             If none, TOL_MERGE will be turned into a digit count and used. 
    
    Returns
    ---------
    hashable:  (n) length array of custom data which can be sorted 
                or used as hash keys
    '''
    as_int   = float_to_int(data, digits)
    dtype    = np.dtype((np.void, as_int.dtype.itemsize * as_int.shape[1]))
    hashable = np.ascontiguousarray(as_int).view(dtype).reshape(-1)
    return hashable

def float_to_int(data, digits=None):
    '''
    Given a numpy array of data represent it as integers.
    '''
    data = np.asanyarray(data)

    dtype_out = np.int32
    if data.size == 0:
        return data.astype(dtype_out)
            
    if digits is None: 
        digits = decimal_to_digits(tol.merge)
    elif isinstance(digits, float) or isinstance(digits, np.float):
        digits = decimal_to_digits(digits)
    elif not (isinstance(digits, int) or isinstance(digits, np.integer)):
        log.warn('Digits were passed as %s!', digits.__class__.__name__)
        raise ValueError('Digits must be None, int, or float!')

    #if data is already an integer or boolean, we're done
    if data.dtype.kind in 'ib':
        as_int = data.astype(dtype_out)
    else:
        data_max = np.abs(data).max() * 10**digits
        dtype_out = [np.int32, np.int64][int(data_max > 2**31)]
        as_int = (np.around(data, digits) * (10**digits)).astype(dtype_out)
    
    return as_int

def unique_ordered(data):
    '''
    Returns the same as np.unique, but ordered as per the
    first occurance of the unique value in data.

    Example
    ---------
    In [1]: a = [0, 3, 3, 4, 1, 3, 0, 3, 2, 1]

    In [2]: np.unique(a)
    Out[2]: array([0, 1, 2, 3, 4])

    In [3]: trimesh.grouping.unique_ordered(a)
    Out[3]: array([0, 3, 4, 1, 2])
    '''
    data   = np.asanyarray(data)
    order  = np.sort(np.unique(data, return_index=True)[1])
    result = data[order]
    return result

def unique_float(data, 
                 return_index   = False,
                 return_inverse = False,
                 digits         = None):
    '''
    Identical to the numpy.unique command, except evaluates floating point 
    numbers, using a specified number of digits. 

    If digits isn't specified, the libray default TOL_MERGE will be used. 
    '''

    as_int = float_to_int(data, digits)
    _junk, unique, inverse = np.unique(as_int, 
                                       return_index   = True,
                                       return_inverse = True)
    if (not return_index) and (not return_inverse):
        return data[unique]
    result = [data[unique]]
    if return_index:   result.append(unique)
    if return_inverse: result.append(inverse)
    return tuple(result)

def unique_rows(data, digits=None):
    '''
    Returns indices of unique rows. It will return the 
    first occurrence of a row that is duplicated:
    [[1,2], [3,4], [1,2]] will return [0,1]

    Arguments
    ---------
    data: (n,m) set of floating point data
    digits: how many digits to consider for the purposes of uniqueness

    Returns
    --------
    unique:  (j) array, index in data which is a unique row
    inverse: (n) length array to reconstruct original
                 example: unique[inverse] == data
    '''
    hashes = hashable_rows(data, digits=digits)
    garbage, unique, inverse = np.unique(hashes, 
                                         return_index   = True, 
                                         return_inverse = True)
    return unique, inverse

def group_rows(data, require_count = None, digits = None):
    '''
    Returns index groups of duplicate rows, for example:
    [[1,2], [3,4], [1,2]] will return [[0,2], [1]]
    
    Arguments
    ----------
    data:          (n,m) array
    require_count: only returns groups of a specified length, eg:
                   require_count =  2
                   [[1,2], [3,4], [1,2]] will return [[0,2]]
    
                   Note that using require_count allows numpy advanced indexing
                   to be used in place of looping and checking hashes, and as a
                   consequence is ~10x faster. 
                   
    digits:        If data is floating point, how many decimals to look at.
                   If this is None, the value in TOL_MERGE will be turned into a 
                   digit count and used. 

    Returns
    ----------
    groups:        List or sequence of indices from data indicating identical rows.
                   If require_count != None, shape will be (j, require_count)
                   If require_count is None, shape will be irregular (AKA a sequence)
    '''
    
    def group_dict():
        '''
        Simple hash table based grouping. 
        The loop and appends make this rather slow on very large arrays,
        But it works on irregular groups nicely, unlike the slicing version of this function
        '''
        observed = dict()
        hashable = hashable_rows(data, digits=digits)
        for index, key in enumerate(hashable):
            key_string = key.tostring()
            if key_string in observed: observed[key_string].append(index)
            else:                      observed[key_string] = [index]
        return np.array(list(observed.values()))
        
    def group_slice():
        # create a representation of the rows that can be sorted
        hashable = hashable_rows(data, digits=digits)
        # record the order of the rows so we can get the original indices back later
        order    = np.argsort(hashable)
        # but for now, we want our hashes sorted
        hashable = hashable[order]
        # this is checking each neighbour for equality, example: 
        # example: hashable = [1, 1, 1]; dupe = [0, 0]
        dupe     = hashable[1:] != hashable[:-1]
        # we want the first index of a group, so we can slice from that location
        # example: hashable = [0 1 1]; dupe = [1,0]; dupe_idx = [0,1]
        dupe_idx = np.append(0, np.nonzero(dupe)[0] + 1)
        # if you wanted to use this one function to deal with non- regular groups
        # you could use: np.array_split(dupe_idx)
        # this is roughly 3x slower than using the group_dict method above. 
        start_ok   = np.diff(np.hstack((dupe_idx, len(hashable)))) == require_count
        groups     = np.tile(dupe_idx[start_ok].reshape((-1,1)), 
                             require_count) + np.arange(require_count)
        groups_idx = order[groups]
        if require_count == 1: 
            return groups_idx.reshape(-1)
        return groups_idx

    if require_count is None: return group_dict()
    else:                     return group_slice()

def boolean_rows(a, b, operation=set.intersection):
    '''
    Find the rows in two arrays which occur in both rows. 

    Arguments
    ---------
    a: (n,d) array
    b: (m,d) array
    operation: boolean set operation function:
               set.intersection
               set.difference
               set.union

    Returns
    --------
    shared: (p, d) array containing rows in both a and b
    '''
    a = float_to_int(a)
    b = float_to_int(b)
    shared = operation({tuple(i) for i in a}, {tuple(j) for j in b})
    shared = np.array(list(shared))
    return shared

def group_vectors(vectors, 
                  angle = np.radians(10), 
                  include_negative = False):
    '''
    Group vectors based on an angle tolerance, with the option to 
    include negative vectors. 
    
    This is very similar to a group_rows(stack_negative(rows))
    The main difference is that max_angle can be much looser, as we
    are doing actual distance queries. 
    '''
    dist_max            = np.tan(angle)
    unit_vectors, valid = unitize(vectors, check_valid = True)
    valid_index         = np.nonzero(valid)[0]
    consumed            = np.zeros(len(unit_vectors), dtype=np.bool)
    tree                = KDTree(unit_vectors)
    unique_vectors      = deque()
    aligned_index       = deque()
    
    for index, vector in enumerate(unit_vectors):
        if consumed[index]: 
            continue
        aligned = np.array(tree.query_ball_point(vector, dist_max))
        vectors = unit_vectors[aligned]
        if include_negative:
            aligned_neg = tree.query_ball_point(-1.0*vector, dist_max)
            vectors     = np.vstack((vectors, -unit_vectors[aligned_neg]))
            aligned     = np.append(aligned, aligned_neg)
        aligned = aligned.astype(int)
        consumed[aligned] = True
        unique_vectors.append(np.median(vectors, axis=0))
        aligned_index.append(valid_index[aligned])
    return np.array(unique_vectors), np.array(aligned_index)

def group_vectors_spherical(vectors, 
                            angle = np.radians(10)):
    '''
    Group vectors based on an angle tolerance, with the option to 
    include negative vectors. 
    
    This is very similar to a group_rows(stack_negative(rows))
    The main difference is that max_angle can be much looser, as we
    are doing actual distance queries. 
    '''
    spherical = vector_to_spherical(vectors)
    angles, groups = group_distance(spherical, angle)
    new_vectors = spherical_to_vector(angles)
    return new_vectors, groups

def group_distance(values, distance):
    consumed = np.zeros(len(values), dtype=np.bool)
    tree     = KDTree(values)

    # (n, d) set of values that are unique
    unique  = deque()
    # (n) sequence of indicies in values
    groups = deque()
    
    for index, value in enumerate(values):
        if consumed[index]: 
            continue
        group = np.array(tree.query_ball_point(value, distance), dtype=np.int)
        consumed[group] = True
        unique.append(np.median(values[group], axis=0))
        groups.append(group)
    return np.array(unique), np.array(groups)
 
def stack_negative(rows):
    '''
    Given an input of rows (n,d), return an array which is (n,2*d)
    Which is sign- independent
    '''
    rows     = np.asanyarray(rows)
    width    = rows.shape[1]
    stacked  = np.column_stack((rows, rows*-1))

    nonzero  = np.abs(rows) > tol.zero
    negative = rows < -tol.zero

    roll = np.logical_and(nonzero, negative).any(axis=1)

    stacked[roll] = np.roll(stacked[roll], 3, axis=1)
    return stacked
            
def clusters(points, radius):
    '''
    Find clusters of points which have neighbours closer than radius
    
    Arguments
    ---------
    points: (n, d) points (of dimension d)
    radius: max distance between points in a cluster

    Returns:
    groups: (m) sequence of indices for points

    '''
    tree   = KDTree(points)
    pairs  = tree.query_pairs(radius)
    graph  = from_edgelist(pairs)
    groups = list(connected_components(graph))
    return groups
                  
def blocks(data, min_len=2, max_len=np.inf, digits=None):
    '''
    Given an array, find the indices of contiguous blocks
    of equal values.

    Arguments
    ---------
    data:    (n) array
    min_len: int, the minimum length group to be returned
    max_len: int, the maximum length group to be retuurned
    digits:  if dealing with floats, how many digits to use

    Returns
    ---------
    blocks: (m) sequence of indices referencing data
    '''
    data = float_to_int(data, digits=digits)

    # find the inflection points, or locations where the array turns
    # from True to False. 
    infl = np.hstack(([0],
                      np.nonzero(np.diff(data))[0] + 1,
                      [len(data)]))
    infl_len = np.diff(infl)    
    infl_ok  = np.logical_and(infl_len >= min_len,
                              infl_len <= max_len)
    if data.dtype.kind == 'b':
        # check to make sure the values of each contiguous block are True,
        # by checking the first value of each block
        infl_ok  = np.logical_and(infl_ok, 
                              data[infl[:-1]])
    # inflate start/end indexes into full ranges of values 
    blocks = [np.arange(infl[i], infl[i+1]) for i, ok in enumerate(infl_ok) if ok]
    return blocks



'''
entities.py: basic geometric primitives

Design intent: only store references to vertex indices and pass the vertex
               array back to functions that require it. 
               This keeps all vertices in one external list.
'''

import numpy as np

from .arc   import discretize_arc, arc_center
from .curve import discretize_bezier, discretize_bspline
from ..util import replace_references

_HASH_LENGTH = 5

class Entity(object):
    def __init__(self, 
                 points, 
                 closed = False):
        self.points = np.array(points)
        self.closed = closed
        
    @property
    def _class_id(self):
        '''
        Return an integer that is unique to the class type.
        Note that this implementation will fail if a class is defined
        that starts with the same letter as an existing class. 
        Since this function is called a lot, it is a tradeoff between 
        speed and robustness where speed won. 
        '''
        return ord(self.__class__.__name__[0])

    @property
    def hash(self):
        '''
        Returns a string unique to the entity.
        If two identical entities exist, they can be removed
        by comparing the string returned by this function.
        '''
        hash = np.zeros(_HASH_LENGTH, dtype=np.int)
        hash[-2:] = self._class_id, int(self.closed)
        points_count = np.min([3, len(self.points)])
        hash[0:points_count] = np.sort(self.points)[-points_count:]
        return hash
        
    def to_dict(self):
        '''
        Returns a dictionary with all of the information about the entity. 
        '''
        return {'type'  : self.__class__.__name__, 
                'points': self.points.tolist(),
                'closed': self.closed}
                
    def rereference(self, replacement):
        '''
        Given a replacement dictionary, change points to reflect the dictionary.
        eg, if replacement = {0:107}, self.points = [0,1902] becomes [107, 1902]
        '''
        self.points = replace_references(self.points, replacement)
        
    @property
    def nodes(self):
        '''
        Returns an (n,2) list of nodes, or vertices on the path.
        Note that this generic class function assumes that all of the reference 
        points are on the path, which is true for lines and three point arcs. 
        If you were to define another class where that wasn't the case
        (for example, the control points of a bezier curve), 
        you would need to implement an entity- specific version of this function.
        
        The purpose of having a list of nodes is so that they can then be added 
        as edges to a graph, so we can use functions to check connectivity, 
        extract paths, etc.  
        
        The slicing on this function is essentially just tiling points
        so the first and last vertices aren't repeated. Example:
        
        self.points = [0,1,2]
        returns:      [[0,1], [1,2]]
        '''
        return np.column_stack((self.points,
                                self.points)).reshape(-1)[1:-1].reshape((-1,2))

    @property
    def end_points(self):
        '''
        Returns the first and last points. Also note that if you
        define a new entity class where the first and last vertices
        in self.points aren't the endpoints of the curve you need to 
        implement this function for your class.
        
        self.points = [0,1,2]
        returns:      [0,2]
        '''
        return self.points[[0,-1]]
            
    @property
    def is_valid(self):
        return True
        
    def reverse(self, direction=-1):
        '''
        Reverse the current entity.
        '''
        self.points = self.points[::direction]

class Line(Entity):
    '''
    A line or poly-line entity
    '''
    def discrete(self, vertices, scale=1.0):
        return vertices[self.points]

    @property
    def closed(self):
        '''
        If the first point is the same as the end point, the polyline is closed
        '''
        ends = self.points[0] == self.points[-1]
        closed = ends and len(self.points) > 2
        return closed

    @closed.setter
    def closed(self, value):
        '''
        Since we can determine easily whether the set of points are closed
        or not, ignore any attempt to manually define the closed property. 
        '''
        return

    @property
    def is_valid(self):
        valid = np.any((self.points - self.points[0]) != 0)
        return valid

class Arc(Entity):
    def discrete(self, vertices, scale=1.0):
        return discretize_arc(vertices[self.points], 
                              close = self.closed,
                              scale = scale)
    def center(self, vertices):
        return arc_center(vertices[self.points])

class Curve(Entity):
    @property
    def _class_id(self):
        return sum([ord(i) for i in self.__class__.__name__])

    @property
    def nodes(self):
        return [[self.points[0], 
                 self.points[1]],
                [self.points[1], 
                 self.points[-1]]]

class Bezier(Curve):
    def discrete(self, vertices, scale=1.0):
        return discretize_bezier(vertices[self.points], scale=scale)

class BSpline(Curve):
    def __init__(self, points, knots, closed=False):
        self.points = points
        self.knots  = knots
        self.closed = closed

    def discrete(self, vertices, count=None, scale=1.0):
        result = discretize_bspline(control = vertices[self.points], 
                                    knots   = self.knots,
                                    count   = count,
                                    scale   = scale)
        return result

    def to_dict(self):
        '''
        Returns a dictionary with all of the information about the entity. 
        '''
        return {'type'  : self.__class__.__name__, 
                'points': self.points.tolist(),
                'knots' : self.knots.tolist(),
                'closed': self.closed}



'''
trimesh.py

Library for importing and doing simple operations on triangular meshes.
'''

import numpy as np

from .points import project_to_plane

def convex_hull(mesh, clean=True):
    '''
    Get a new Trimesh object representing the convex hull of the 
    current mesh. Requires scipy >.12.
    
    Argments
    --------
    clean: boolean, if True will fix normals and winding
    to be coherent (as qhull/scipy outputs are not)

    Returns
    --------
    convex: Trimesh object of convex hull of current mesh
    '''
    from scipy.spatial import ConvexHull

    faces  = ConvexHull(mesh.vertices.view(np.ndarray)).simplices
    convex = mesh.__class__(vertices = mesh.vertices.view(np.ndarray).copy(), 
                            faces    = faces,
                            process  = clean)
    if clean:
        # the normals and triangle winding returned by scipy/qhull's
        # ConvexHull are apparently random, so we need to completely fix them
        convex.fix_normals()
        # since we just copied all the vertices over, we will have a bunch
        # of unreferenced vertices, so it is best to remove them
        convex.remove_unreferenced_vertices()
    return convex

def is_convex(mesh):
    '''
    Test whether a mesh is convex or not.
    '''
    test = convex_hull(mesh, clean=False)
    convex = test.faces.shape == mesh.faces.shape
    return convex
    
def planar_hull(vertices,
                normal, 
                origin           = [0,0,0], 
                return_transform = False):
    planar , T = project_to_plane(vertices,
                                  plane_normal     = normal,
                                  plane_origin     = origin,
                                  return_transform = True)
    hull_edges = ConvexHull(planar).simplices
    if return_transform:
        return planar[hull_edges], T
    return planar[hull_edges]



'''
Functions dealing with (n,d) points
'''
import numpy as np

from .constants import log, tol
from .geometry  import plane_transform
from .util      import unitize, TrackedArray, Cache
    
def transform_points(points, matrix, translate=True):
    '''
    Returns points, rotated by transformation matrix 
    If points is (n,2), matrix must be (3,3)
    if points is (n,3), matrix must be (4,4)

    Arguments
    ----------
    points: (n, 2|3) set of points
    matrix: (3|4, 3|4) rotation matrix
    translate: boolean, apply translation from matrix or not

    '''
    dimension   = np.shape(points)[1]
    column      = np.zeros(len(points)) + int(bool(translate))
    stacked     = np.column_stack((points, column))
    transformed = np.dot(matrix, stacked.T).T[:,0:dimension]
    return transformed

def point_plane_distance(points, plane_normal, plane_origin=[0,0,0]):
    w         = np.array(points) - plane_origin
    distances = np.dot(plane_normal, w.T) / np.linalg.norm(plane_normal)
    return distances


def major_axis(points):
    '''
    Returns an approximate vector representing the major axis of points
    '''
    u,s,v = np.linalg.svd(points)
    axis = v[0]
    return axis 

def surface_normal(points):
    '''
    Returns a normal estimate of a group of points using SVD

    Arguments
    ---------
    points: (n,d) set of points

    Returns
    ---------
    normal: (d) vector
    '''
    normal = np.linalg.svd(points)[2][-1]
    return normal

def plane_fit(points, tolerance=None):
    '''                                   
    Given a set of points, find an origin and normal using least squares
    Arguments
    ---------
    points: (n,3)
    tolerance: how non-planar the result can be without raising an error
    
    Returns
    ---------
    C: (3) point on the plane
    N: (3) normal vector
    '''

    C = points[0]
    x = points - C
    M = np.dot(x.T, x)
    N = np.linalg.svd(M)[0][:,-1]

    if not (tolerance is None):
        normal_range  = np.ptp(np.dot(N, points.T))
        if normal_range > tol.planar:
            log.error('Points have peak to peak of %f', normal_range)
            raise ValueError('Plane outside tolerance!')
    return C, N

def radial_sort(points, 
                origin = None, 
                normal = None):
    '''
    Sorts a set of points radially (by angle) around an origin/normal.
    If origin/normal aren't specified, it sorts around centroid
    and the approximate plane the points lie in. 

    points: (n,3) set of points
    '''
    #if origin and normal aren't specified, generate one at the centroid
    if origin==None: origin = np.average(points, axis=0)
    if normal==None: normal = surface_normal(points)
    
    # create two axis perpendicular to each other and the normal, 
    # and project the points onto them
    axis0 = [normal[0], normal[2], -normal[1]]
    axis1 = np.cross(normal, axis0)
    ptVec = points - origin
    pr0 = np.dot(ptVec, axis0)
    pr1 = np.dot(ptVec, axis1)

    #calculate the angles of the points on the axis
    angles = np.arctan2(pr0, pr1)

    #return the points sorted by angle
    return points[[np.argsort(angles)]]
               
def project_to_plane(points, 
                     plane_normal     = [0,0,1], 
                     plane_origin     = [0,0,0],
                     transform        = None,
                     return_transform = False,
                     return_planar    = True):
    '''
    Projects a set of (n,3) points onto a plane.

    Arguments
    ---------
    points:           (n,3) array of points
    plane_normal:     (3) normal vector of plane
    plane_origin:     (3) point on plane
    transform:        None or (4,4) matrix. If specified, normal/origin are ignored
    return_transform: bool, if true returns the (4,4) matrix used to project points 
                      onto a plane
    return_planar:    bool, if True, returns (n,2) points. If False, returns
                      (n,3), where the Z column consists of zeros
    '''

    if np.all(np.abs(plane_normal) < tol.zero):
        raise NameError('Normal must be nonzero!')

    if transform is None:
        transform = plane_transform(plane_origin, plane_normal)
        
    transformed = transform_points(points, transform)[:,0:(3-int(return_planar))]

    if return_transform: 
        polygon_to_3D = np.linalg.inv(transform)
        return transformed, polygon_to_3D
    return transformed

def absolute_orientation(points_A, points_B, return_error=False):
    '''
    Calculates the transform that best aligns points_A with points_B
    Uses Horn's method for the absolute orientation problem, in 3D with no scaling.

    Arguments
    ---------
    points_A:     (n,3) list of points
    points_B:     (n,3) list of points, T*points_A
    return_error: boolean, if True returns (n) list of euclidean distances
                  representing the distance from  T*points_A[i] to points_B[i]


    Returns
    ---------
    M:    (4,4) transformation matrix for the transform that best aligns 
           points_A to points_B
    error: float, list of maximum euclidean distance
    '''

    points_A = np.array(points_A)
    points_B = np.array(points_B)
    if (points_A.shape != points_B.shape):
        raise ValueError('Points must be of the same shape!')
    if len(points_A.shape) != 2 or points_A.shape[1] != 3:
        raise ValueError('Points must be (n,3)!')

    lc = np.average(points_A, axis=0)
    rc = np.average(points_B, axis=0)
    left  = points_A - lc
    right = points_B - rc
    M = np.dot(left.T, right)
    [[Sxx, Sxy, Sxz], 
     [Syx, Syy, Syz], 
     [Szx, Szy, Szz]] = M
    N=[[(Sxx+Syy+Szz), (Syz-Szy), (Szx-Sxz), (Sxy-Syx)],
       [(Syz-Szy), (Sxx-Syy-Szz), (Sxy+Syx), (Szx+Sxz)],
       [(Szx-Sxz), (Sxy+Syx), (-Sxx+Syy-Szz), (Syz+Szy)],
       [(Sxy-Syx), (Szx+Sxz), (Syz+Szy),(-Sxx-Syy+Szz)]]
    (w,v) = np.linalg.eig(N)
    q = v[:,np.argmax(w)]
    q = q/np.linalg.norm(q)
    M1 = [[q[0], -q[1], -q[2], -q[3]], 
          [q[1],  q[0],  q[3], -q[2]], 
          [q[2], -q[3],  q[0],  q[1]],
          [q[3],  q[2], -q[1],  q[0]]]
    M2 = [[q[0], -q[1], -q[2], -q[3]], 
          [q[1],  q[0], -q[3],  q[2]], 
          [q[2],  q[3],  q[0], -q[1]],
          [q[3], -q[2],  q[1],  q[0]]]
    R = np.dot(np.transpose(M1),M2)[1:4,1:4]
    T = rc - np.dot(R, lc)
    
    M          = np.eye(4) 
    M[0:3,0:3] = R
    M[0:3,3]   = T

    if return_error:
        errors = np.sum((transform_points(points_A, M) - points_B)**2, axis = 1)
        return M, errors.max()
    return M

def remove_close(points, radius):
    '''
    Given an (n, m) set of points where n=(2|3) return a list of points
    where no point is closer than radius
    '''
    from scipy.spatial import cKDTree as KDTree

    tree     = KDTree(points)
    consumed = np.zeros(len(points), dtype=np.bool)
    unique   = np.zeros(len(points), dtype=np.bool)
    for i in xrange(len(points)):
        if consumed[i]: continue
        neighbors = tree.query_ball_point(points[i], r=radius)
        consumed[neighbors] = True
        unique[i]           = True
    return points[unique]

def remove_close_set(points_fixed, points_reduce, radius):
    '''
    Given two sets of points and a radius, return a set of points
    that is the subset of points_reduce where no point is within 
    radius of any point in points_fixed
    '''
    from scipy.spatial import cKDTree as KDTree

    tree_fixed  = KDTree(points_fixed)
    tree_reduce = KDTree(points_reduce)
    reduce_duplicates = tree_fixed.query_ball_tree(tree_reduce, r = radius)
    reduce_duplicates = np.unique(np.hstack(reduce_duplicates).astype(int))
    reduce_mask = np.ones(len(points_reduce), dtype=np.bool)
    reduce_mask[reduce_duplicates] = False
    points_clean = points_reduce[reduce_mask]
    return points_clean
    
def k_means(points, k, **kwargs):
    '''
    Find k centroids that attempt to minimize the k- means problem:
    https://en.wikipedia.org/wiki/Metric_k-center
    
    Arguments
    ----------
    points: (n, d) set of points
    k: int, number of centroids to compute
    **kwargs: passed directly to scipy.cluster.vq.kmeans
    
    Returns
    ----------
    centroids: (k, d) set of points
    labels: (n) set of indexes for which points belong to which centroid
    '''
    from scipy.cluster.vq import kmeans
    from scipy.spatial import cKDTree

    points = np.asanyarray(points)
    points_std = points.std(axis=0)
    whitened = points / points_std
    centroids_whitened, distortion = kmeans(whitened, k, **kwargs)
    centroids = centroids_whitened * points_std
    tree = cKDTree(centroids)
    labels = tree.query(points, k=1)[1]
    return centroids, labels
    
def plot_points(points, show=True):    
    import matplotlib.pyplot as plt

    dimension = np.shape(points)[1]    
    if dimension == 3:
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax  = fig.add_subplot(111, projection='3d')
        ax.scatter(*np.array(points).T)
    elif dimension == 2:
        plt.scatter(*np.array(points).T)
    else:
        raise ValueError('Points must be 2D or 3D, not %dD', dimension)

    if show: plt.show()

class PointCloud(TrackedArray):
    '''
    Subclass of TrackedArray, which is subclass of np.ndarray
    '''
    def show(self):
        plot_points(self)




import numpy as np
from collections import OrderedDict

def load_ply(file_obj, *args, **kwargs):
    '''
    Load a PLY file from an open file object.
    
    Arguments
    ---------
    file_obj: an open file- like object

    Returns
    ---------
    mesh_kwargs: dictionary of mesh info which can be passed to 
                 Trimesh constructor, eg: a = Trimesh(**mesh_kwargs)
    '''
    # OrderedDict which is populated from the header
    elements = ply_read_header(file_obj)
    # some elements are passed where the list dimensions
    # are not included in the header, so this function goes 
    # into the meat of the file and grabs the list dimensions 
    # before we to the main data read as a single operation
    ply_populate_listsize(file_obj, elements)

    # how many bytes are left in the file
    size_file = size_to_end(file_obj)
    # how many bytes should the data structure described by
    # the header take up
    size_elements = ply_elements_size(elements)
    
    # if the number of bytes is not the same the file is probably corrupt
    if size_file != size_elements:
        raise ValueError('File is unexpected length!')

    # with everything populated and a reasonable confidence the file
    # is intact, read the data fields described by the header
    ply_populate_data(file_obj, elements)
    # all of the data is now stored in elements, but we need it as
    # a set of keyword arguments we can pass to the Trimesh constructor
    # will look something like {'vertices' : (data), 'faces' : (data)} 
    mesh_kwargs = ply_elements_kwargs(elements)
    return mesh_kwargs

def ply_element_colors(element):
    '''
    Given an element, try to extract RGBA color from its properties
    and return them as an (n,3|4) array.
    '''
    color_keys = ['red', 'green', 'blue', 'alpha']
    candidate_colors = [element['data'][i] for i in color_keys if i in element['properties']]

    if len(candidate_colors) >= 3:
        return np.column_stack(candidate_colors)
    return None

def ply_read_header(file_obj):
    '''
    Read the ASCII header of a PLY file, and leave the file object 
    at the position of the start of data. 
    '''
    # from ply specification
    dtypes = {'char'  : 'i1',
              'uchar' : 'u1',
              'short' : 'i2',
              'ushort': 'u2',
              'int'   : 'i4',
              'uint'  : 'u4',
              'float' : 'f4',
              'double': 'f8'}

    if not 'ply' in str(file_obj.readline()):
        raise ValueError('This aint a ply file')
    encoding = str(file_obj.readline()).strip().split()[1]

    if 'ascii' in encoding:
        raise ValueError('ASCII PLY not supported!')

    endian   = ['<', '>']['big' in encoding]
    elements = OrderedDict()

    while True:
        line = file_obj.readline()
        if line is None:
            raise ValueError('Header wasn\'t terminated properly!')
        line = line.decode('utf-8').strip().split()
        
        if 'end_header' in line:
            break

        if 'element' in line[0]:
            name, length = line[1:]
            elements[name] = {'length'     : int(length), 
                              'properties' : OrderedDict()}
        elif 'property' in line[0]:
            if len(line) == 3:
                dtype, field = line[1:]
                elements[name]['properties'][str(field)] = endian + dtypes[dtype]
            elif 'list' in line[1]:
                dtype_count, dtype, field = line[2:]
                elements[name]['properties'][str(field)] = (endian +
                                                            dtypes[dtype_count] + 
                                                            ', ($LIST,)'+
                                                            endian + 
                                                            dtypes[dtype])
    return elements

def ply_populate_listsize(file_obj, elements):
    '''
    Given a set of elements populated from the header if there are any
    list properties seek in the file the length of the list. 

    Note that if you have a list where each instance is different length
    (if for example you mixed triangles and quads) this won't work at all
    '''
    p_start = file_obj.tell()
    p_current = file_obj.tell()
    for element_key, element in elements.items():
        props = element['properties']
        prior_data = ''
        for k, dtype in props.items():
            if '$LIST' in dtype:
                # every list field has two data types: 
                # the list length (single value), and the list data (multiple)
                # here we are only reading the single value for list length
                field_dtype = np.dtype(dtype.split(',')[0])
                if len(prior_data) == 0: 
                    offset = 0
                else: 
                    offset = np.dtype(prior_data).itemsize()
                file_obj.seek(p_current+offset)
                size = np.fromstring(file_obj.read(field_dtype.itemsize), 
                                     dtype=field_dtype)[0]
                props[k] = props[k].replace('$LIST', str(size))
            prior_data += props[k] +','
        itemsize = np.dtype(', '.join(props.values())).itemsize
        p_current += element['length'] * itemsize
    file_obj.seek(p_start)
    
def ply_populate_data(file_obj, elements):
    '''
    Given the data type and field information from the header,
    read the data and add it to a 'data' field in the element.
    '''
    for key in elements.keys():
        items = list(elements[key]['properties'].items())
        dtype = np.dtype(items)
        data  = file_obj.read(elements[key]['length'] * dtype.itemsize)
        elements[key]['data'] = np.fromstring(data, dtype=dtype)
    return elements

def ply_elements_kwargs(elements):
    '''
    Given an elements data structure, extract the keyword
    arguments that a Trimesh object constructor will expect.
    '''
    vertices = np.column_stack([elements['vertex']['data'][i] for i in 'xyz'])
    faces    = elements['face']['data']['vertex_indices']['f1']
    face_colors   = ply_element_colors(elements['face'])
    vertex_colors = ply_element_colors(elements['vertex'])
    result = {'vertices' : vertices,
              'faces'    : faces,
              'face_colors'   : face_colors,
              'vertex_colors' : vertex_colors}
    return result

def ply_elements_size(elements):
    '''
    Given an elements data structure populated from the header, 
    calculate how long the file should be if it is intact.
    '''
    size = 0
    for element in elements.values():
        dtype = np.dtype(','.join(element['properties'].values()))
        size += element['length'] * dtype.itemsize
    return size

def size_to_end(file_obj):
    '''
    Given an open file object, return the number of bytes 
    to the end of the file
    '''
    position_current = file_obj.tell()
    file_obj.seek(0,2)
    position_end = file_obj.tell()
    file_obj.seek(position_current)
    size = position_end - position_current
    return size

_ply_loaders = {'ply' : load_ply}



import numpy as np
from .grouping   import unique_float
from .points     import plot_points
from collections import deque

class Voxel:
    def __init__(self, mesh, pitch): 
        self._run = mesh_to_run(mesh, pitch)
        self._raw = None
        
    @property
    def raw(self):
        '''
        Generate a raw 3D boolean array from the internal run-length encoded data
        '''
        if self._raw is  None:
            self._raw = run_to_raw(**self.run)
        return self._raw
         
    @property
    def run(self):
        return self._run
        
    @property
    def pitch(self):
        return self.run['pitch']

    @property
    def origin(self):
        return self.run['origin']

    def volume(self):
        volume = self.raw.sum() * (self.pitch**2)
        return volume
        
    def show(self):
        plot_raw(self.raw, **self.run)

def run_to_raw(shape, index_xy, index_z, **kwargs):
    raw = np.zeros(shape, dtype=np.bool)
    for xy, z in zip(index_xy, index_z):
        for z_start, z_end in np.reshape(z,(-1,2)):
            raw[xy[0], xy[1]][z_start:z_end] = True
    return raw

def mesh_to_run(mesh, pitch):
    '''
    Convert a mesh to a run-length encoded voxel grid. 

    This is done via ray tests which return intersection points,
    which is easily convertable to a raw 3D boolean voxel array
    '''
    bounds    = mesh.bounds / pitch
    bounds[0] = np.floor(bounds[0]) * pitch
    bounds[1] = np.ceil( bounds[1]) * pitch
    
    x_grid = np.arange(*bounds[:,0], step = pitch)
    y_grid = np.arange(*bounds[:,1], step = pitch)    
    grid   = np.dstack(np.meshgrid(x_grid, y_grid)).reshape((-1,2))

    ray_origins  = np.column_stack((grid, np.tile(bounds[0][2], len(grid))))
    ray_origins += [pitch*.5, pitch*.5, -pitch]
    ray_vectors  = np.tile([0.0,0.0,1.0], (len(grid),1))
    rays         = np.column_stack((ray_origins, 
                                    ray_vectors)).reshape((-1,2,3))

    hits        = mesh.ray.intersects_location(rays)    
    raw_shape   = np.ptp(bounds/pitch, axis=0).astype(int)
    grid_origin = bounds[0]
    grid_index  = ((grid/pitch) - (grid_origin[0:2]/pitch)).astype(int)

    run_z  = deque()
    run_xy = deque()

    for i, hit in enumerate(hits):
        if len(hit) == 0: continue

        z = hit[:,2]-grid_origin[2]
        # if a ray hits exactly on an edge, there will
        # be a duplicate entry in hit (for both triangles) 
        z = unique_float(z)

        index_z = np.round(z/pitch).astype(int)
        # if a hit is on edge, it will be returned twice
        # this np.unique call returns sorted, unique indicies
        index_z.sort()

        if np.mod(len(index_z), 2) != 0:
            # this is likely the on-vertex case
            index_z = index_z[[0,-1]]

        run_z.append(index_z)
        run_xy.append(grid_index[i])

    result = {'shape'    : raw_shape, 
              'index_xy' : np.array(run_xy), 
              'index_z'  : np.array(run_z), 
              'origin'   : grid_origin,
              'pitch'    : pitch}    
    return result        

def plot_raw(raw, pitch, origin, **kwargs):
    render = np.column_stack(np.nonzero(raw))*pitch + origin
    plot_points(render)



import numpy as np

from ..base      import Trimesh
from ..constants import _log_time, log
from ..util      import is_file, is_dict, is_string, make_sequence, is_instance_named

from .assimp import _assimp_loaders
from .stl    import _stl_loaders
from .misc   import _misc_loaders
from .step   import _step_loaders
from .ply    import _ply_loaders

def load_path(*args, **kwargs):
    raise ImportError('No path functionality available!')
def path_formats():
    return []

try:
    from ..path.io.load import load_path, path_formats
except:
    log.warning('No path functionality available!', exc_info=True)

def mesh_formats():
    return list(_mesh_loaders.keys())

def available_formats():
    return np.append(mesh_formats(), path_formats())

def load(obj, file_type=None, **kwargs):
    '''
    Load a mesh or vectorized path into a 
    Trimesh, Path2D, or Path3D object.

    Arguments
    ---------
    file_obj: a filename string or a file-like object
    file_type: str representing file type (eg: 'stl')

    Returns:
    geometry: Trimesh, Path2D, Path3D, or list of same. 
    '''

    if isinstance(obj, Trimesh):
        return obj

    if is_instance_named(obj, 'Path'):
        return obj

    if is_string(obj):
        file_type = (str(obj).split('.')[-1]).lower()
        obj = open(obj, 'rb')
        
    if file_type is None:
        file_type = obj.__class__.__name__

    if file_type in path_formats():
        return load_path(obj, file_type, **kwargs)
    elif file_type in mesh_formats():
        return load_mesh(obj, file_type, **kwargs)
        
    raise ValueError('File type: %s not supported', str(file_type))

@_log_time
def load_mesh(obj, file_type=None, process=True):
    '''
    Load a mesh file into a Trimesh object

    Arguments
    ---------
    file_obj: a filename string or a file-like object
    file_type: str representing file type (eg: 'stl')
    process:   boolean flag, whether to process the mesh on load

    Returns:
    mesh: a single Trimesh object, or a list of Trimesh objects, 
          depending on the file format. 
    
    '''

    if is_string(obj):
        file_type = (str(obj).split('.')[-1]).lower()
        obj = open(obj, 'rb')

    if file_type is None:
        file_type = obj.__class__.__name__
    
    file_type = str(file_type).lower()
    loaded = _mesh_loaders[file_type](obj, file_type)
    
    if is_file(obj): 
        obj.close()
    
    log.debug('loaded mesh using %s',
              _mesh_loaders[file_type].__name__)

    meshes = [Trimesh(process=process, **i) for i in make_sequence(loaded)]
    
    if len(meshes) == 1: 
        return meshes[0]
    return meshes

_mesh_loaders = {}
# assimp has a lot of loaders, but they are all quite slow
# so we load them first and replace them with native loaders if possible
_mesh_loaders.update(_assimp_loaders)
_mesh_loaders.update(_stl_loaders)
_mesh_loaders.update(_misc_loaders)
_mesh_loaders.update(_step_loaders)
_mesh_loaders.update(_ply_loaders)



import numpy as np

from .points    import unitize, point_plane_distance
from .constants import tol

def cross(triangles):
    '''
    Returns the cross product of two edges from input triangles 

    triangles: vertices of triangles (n,3,3)
    returns:   cross product of two edge vectors (n,3)
    '''
    vectors = np.diff(triangles, axis=1)
    crosses = np.cross(vectors[:,0], vectors[:,1])
    return crosses
    
def area(triangles, sum=True):
    '''
    Calculates the sum area of input triangles 

    triangles: vertices of triangles (n,3,3)
    returns:   area, (n)
    '''
    crosses = cross(triangles)
    area    = (np.sum(crosses**2, axis=1)**.5)*.5
    if sum: 
        return np.sum(area)
    return area
    
def normals(triangles):
    '''
    Calculates the normals of input triangles 
    
    triangles: vertices of triangles, (n,3,3)
    returns:   normal vectors, (n,3)
    '''
    crosses = cross(triangles)
    normals, valid = unitize(crosses, check_valid=True)
    return normals, valid
    
def all_coplanar(triangles):
    '''
    Given a list of triangles, return True if they are all coplanar, and False if not.
  
    triangles: vertices of triangles, (n,3,3)
    returns:   all_coplanar, bool
    '''
    test_normal  = normal(triangles)[0]
    test_vertex  = triangles[0][0]
    distances    = point_plane_distance(points       = triangles[1:].reshape((-1,3)),
                                        plane_normal = test_normal,
                                        plane_origin = test_vertex)
    all_coplanar = np.all(np.abs(distances) < tol.zero)
    return all_coplanar
    
def any_coplanar(triangles):
    '''
    Given a list of triangles, if the FIRST triangle is coplanar with ANY
    of the following triangles, return True.
    Otherwise, return False. 
    '''
    test_normal  = normals(triangles)[0]
    test_vertex  = triangles[0][0]
    distances    = point_plane_distance(points       = triangles[1:].reshape((-1,3)),
                                        plane_normal = test_normal,
                                        plane_origin = test_vertex)
    any_coplanar = np.any(np.all(np.abs(distances.reshape((-1,3)) < tol.zero), axis=1))
    return any_coplanar
    
def mass_properties(triangles, density = 1.0, skip_inertia=False):
    '''
    Calculate the mass properties of a group of triangles.
    
    Implemented from:
    http://www.geometrictools.com/Documentation/PolyhedralMassProperties.pdf
    '''
    crosses      = cross(triangles)
    surface_area = np.sum(np.sum(crosses**2, axis=1)**.5)*.5

    # these are the subexpressions of the integral 
    f1 = triangles.sum(axis=1)
    
    # triangles[:,0,:] will give rows like [[x0, y0, z0], ...] (the first vertex of every triangle)
    # triangles[:,:,0] will give rows like [[x0, x1, x2], ...] (the x coordinates of every triangle)
    f2 = (triangles[:,0,:]**2 +
          triangles[:,1,:]**2 +    
          triangles[:,0,:]*triangles[:,1,:] + 
          triangles[:,2,:]*f1)
          
    f3 = ((triangles[:,0,:]**3) + 
          (triangles[:,0,:]**2) * (triangles[:,1,:]) + 
          (triangles[:,0,:])    * (triangles[:,1,:]**2) +
          (triangles[:,1,:]**3) + 
          (triangles[:,2,:]*f2))
          
    g0 = (f2 + (triangles[:,0,:] + f1)*triangles[:,0,:])
    g1 = (f2 + (triangles[:,1,:] + f1)*triangles[:,1,:])
    g2 = (f2 + (triangles[:,2,:] + f1)*triangles[:,2,:])
    
    integral      = np.zeros((10, len(f1)))
    integral[0]   = crosses[:,0] * f1[:,0]
    integral[1:4] = (crosses * f2).T
    integral[4:7] = (crosses * f3).T
    
    for i in range(3):
        triangle_i    = np.mod(i+1, 3)
        integral[i+7] = crosses[:,i] * ((triangles[:,0, triangle_i] * g0[:,i]) + 
                                        (triangles[:,1, triangle_i] * g1[:,i]) + 
                                        (triangles[:,2, triangle_i] * g2[:,i]))
                                        
    coefficents = 1.0 / np.array([6,24,24,24,60,60,60,120,120,120])
    integrated  = integral.sum(axis=1) * coefficents
    
    volume      = integrated[0]
    center_mass = integrated[1:4] / volume

    result = {'density'      : density,
              'surface_area' : surface_area,
              'volume'       : volume,
              'mass'         : density * volume,
              'center_mass'  : center_mass.tolist()}

    if skip_inertia: 
        return result
              
    inertia = np.zeros((3,3))
    inertia[0,0] = integrated[5] + integrated[6] - (volume * (center_mass[[1,2]]**2).sum())
    inertia[1,1] = integrated[4] + integrated[6] - (volume * (center_mass[[0,2]]**2).sum())
    inertia[2,2] = integrated[4] + integrated[5] - (volume * (center_mass[[0,1]]**2).sum())
    inertia[0,1] = (integrated[7] - (volume * np.product(center_mass[[0,1]])))
    inertia[1,2] = (integrated[8] - (volume * np.product(center_mass[[1,2]])))
    inertia[0,2] = (integrated[9] - (volume * np.product(center_mass[[0,2]])))
    inertia[2,0] = inertia[0,2]
    inertia[2,1] = inertia[1,2]
    inertia[1,0] = inertia[0,1]
    inertia *= density
    
    result['inertia'] = inertia.tolist()
    
    return result

def bounds_tree(triangles):
    '''
    Given a set of triangles, create an r-tree for broad- phase 
    collision detection

    Arguments
    ---------
    triangles: (n, 3, 3) list of vertices

    Returns
    ---------
    tree: Rtree object 
    '''
    from rtree import index

    # the property object required to get a 3D r-tree index
    properties = index.Property()
    properties.dimension = 3
    # the (n,6) interleaved bounding box for every triangle
    tri_bounds = np.column_stack((triangles.min(axis=1), triangles.max(axis=1)))
  
    # stream loading wasn't getting proper index
    tree = index.Index(properties=properties)  
    for i, bounds in enumerate(tri_bounds):
        tree.insert(i, bounds)    
    return tree



import numpy as np
import json
import os 

from .dxf_load   import load_dxf
from .svg_load   import svg_to_path
from .misc       import lines_to_path, polygon_to_path, dict_to_path

from ..path      import Path, Path2D, Path3D

from ...constants import log
from ...util      import is_sequence, is_file, is_string, is_instance_named

def load_path(obj, file_type=None):
    '''
    Utility function which can be passed a filename, file object, or list of lines
    '''

    if isinstance(obj, Path):
        return obj
    elif is_file(obj):
        loaded = _LOADERS[file_type](obj)
        obj.close()
    elif is_string(obj):
        file_obj  = open(obj, 'rb')
        file_type = os.path.splitext(obj)[-1][1:].lower()
        loaded = _LOADERS[file_type](file_obj)
        file_obj.close()
    elif is_instance_named(obj, 'Polygon'):
        loaded = polygon_to_path(obj)
    elif is_instance_named(obj, 'dict'):
        loaded = dict_to_path(obj)
    elif is_sequence(obj):
        loaded = lines_to_path(obj)
    else:
        raise ValueError('Not a supported object type!')
    path = _create_path(**loaded)
    return path

def _create_path(entities, vertices, metadata=None):
    shape = np.shape(vertices)
    if ((len(shape) != 2) or 
        (not shape[1] in [2,3])):
        raise ValueError('Vertices must be 2D or 3D!')
    path = [Path2D, Path3D][shape[1] == 3](entities = entities, 
                                           vertices = vertices,
                                           metadata = metadata)
    return path

def path_formats():
    return list(_LOADERS.keys())

_LOADERS = {'dxf': load_dxf,
            'svg': svg_to_path}



'''
trimesh.util: utility functions

Only imports from numpy and the standard library are allowed in this file.
'''

import numpy as np
import time
import logging
import hashlib
import base64

from collections import defaultdict, deque
from sys import version_info

if version_info.major >= 3:
    basestring = str

log = logging.getLogger('trimesh')
log.addHandler(logging.NullHandler())   
    
# included here so util has only standard library imports
_TOL_ZERO = 1e-12

def unitize(points, check_valid=False):
    '''
    Turn a list of vectors into a list of unit vectors.
    
    Arguments
    ---------
    points:       (n,m) or (j) input array of vectors. 
                  For 1D arrays, points is treated as a single vector
                  For 2D arrays, each row is treated as a vector
    check_valid:  boolean, if True enables valid output and checking

    Returns
    ---------
    unit_vectors: (n,m) or (j) length array of unit vectors

    valid:        (n) boolean array, output only if check_valid.
                   True for all valid (nonzero length) vectors, thus m=sum(valid)
    '''
    points = np.asanyarray(points)
    axis   = len(points.shape) - 1
    length = np.sum(points ** 2, axis=axis) ** .5
    if check_valid:
        valid = np.greater(length, _TOL_ZERO)
        if axis == 1: 
            unit_vectors = (points[valid].T / length[valid]).T
        elif len(points.shape) == 1 and valid: 
            unit_vectors = points / length
        else:         
            unit_vectors = np.array([])
        return unit_vectors, valid        
    else: 
        unit_vectors = (points.T / length).T
    return unit_vectors

def transformation_2D(offset=[0.0,0.0], theta=0.0):
    '''
    2D homogeonous transformation matrix
    '''
    T = np.eye(3)
    s = np.sin(theta)
    c = np.cos(theta)

    T[0,0:2] = [ c, s]
    T[1,0:2] = [-s, c]
    T[0:2,2] = offset
    return T

def euclidean(a, b):
    '''
    Euclidean distance between vectors a and b
    '''
    return np.sum((np.array(a) - b)**2) ** .5

def is_file(obj):
    return hasattr(obj, 'read')

def is_string(obj):
    return isinstance(obj, basestring)

def is_dict(obj):
    return isinstance(obj, dict)

def is_sequence(obj):
    '''
    Returns True if obj is a sequence.
    '''
    seq = (not hasattr(obj, "strip") and
           hasattr(obj, "__getitem__") or
           hasattr(obj, "__iter__"))

    seq = seq and not isinstance(obj, dict)
    # numpy sometimes returns objects that are single float64 values
    # but sure look like sequences, so we check the shape
    if hasattr(obj, 'shape'):
        seq = seq and obj.shape != ()
    return seq

def is_shape(obj, shape):
    '''
    Compare the shape of a numpy.ndarray to a target shape, 
    with any value less than zero being considered a wildcard

    Arguments
    ---------
    obj: np.ndarray to check the shape of
    shape: list or tuple of shape. 
           Any negative term will be considered a wildcard
           Any tuple term will be evaluated as an OR

    Returns
    ---------
    shape_ok: bool, True if shape of obj matches query shape

    Examples
    ------------------------
    In [1]: a = np.random.random((100,3))

    In [2]: a.shape
    Out[2]: (100, 3)

    In [3]: trimesh.util.is_shape(a, (-1,3))
    Out[3]: True

    In [4]: trimesh.util.is_shape(a, (-1,3,5))
    Out[4]: False

    In [5]: trimesh.util.is_shape(a, (100,-1))
    Out[5]: True

    In [6]: trimesh.util.is_shape(a, (-1,(3,4)))
    Out[6]: True

    In [7]: trimesh.util.is_shape(a, (-1,(4,5)))
    Out[7]: False
    '''

    if (not hasattr(obj, 'shape') or
        len(obj.shape) != len(shape)):
        return False

    for i, target in zip(obj.shape, shape):
        # check if current field has multiple acceptable values
        if is_sequence(target):
            if i in target: continue
            else:           return False
        # check if current field is a wildcard
        if target < 0: 
            continue
        # since we have a single target and a single value,
        # if they are not equal we have an answer
        if target != i: 
            return False

    # since none of the checks failed, the two shapes are the same
    return True

def make_sequence(obj):
    '''
    Given an object, if it is a sequence return, otherwise
    add it to a length 1 sequence and return.

    Useful for wrapping functions which sometimes return single 
    objects and other times return lists of objects. 
    '''
    if is_sequence(obj): return np.array(obj)
    else:                return np.array([obj])

def vector_to_spherical(cartesian):
    '''
    Convert a set of cartesian points to (n,2) spherical vectors
    '''
    x,y,z = np.array(cartesian).T
    # cheat on divide by zero errors
    x[np.abs(x) < _TOL_ZERO] = _TOL_ZERO
    spherical = np.column_stack((np.arctan(y/x),
                                 np.arccos(z)))
    return spherical

def spherical_to_vector(spherical):
    '''
    Convert a set of (n,2) spherical vectors to (n,3) vectors
    '''
    theta, phi = np.array(spherical).T
    st, ct = np.sin(theta), np.cos(theta)
    sp, cp = np.sin(phi),   np.cos(phi)
    vectors = np.column_stack((ct*sp,
                               st*sp,
                               cp))
    return vectors

def diagonal_dot(a, b):
    '''
    Dot product by row of a and b.

    Same as np.diag(np.dot(a, b.T)) but without the monstrous 
    intermediate matrix.
    '''
    result = (np.array(a)*b).sum(axis=1)
    return result

def three_dimensionalize(points, return_2D=True):
    '''
    Given a set of (n,2) or (n,3) points, return them as (n,3) points

    Arguments
    ----------
    points:    (n, 2) or (n,3) points
    return_2D: boolean flag

    Returns
    ----------
    if return_2D: 
        is_2D: boolean, True if points were (n,2)
        points: (n,3) set of points
    else:
        points: (n,3) set of points
    '''
    points = np.array(points)
    shape = points.shape
 
    if len(shape) != 2:
        raise ValueError('Points must be 2D array!')

    if shape[1] == 2:
        points = np.column_stack((points, np.zeros(len(points))))
        is_2D = True
    elif shape[1] == 3:
        is_2D = False
    else:
        raise ValueError('Points must be (n,2) or (n,3)!')

    if return_2D: 
        return is_2D, points
    return points

def grid_arange_2D(bounds, step):
    '''
    Return a 2D grid with specified spacing

    Arguments
    ---------
    bounds: (2,2) list of [[minx, miny], [maxx, maxy]]
    step:   float, separation between points
    
    Returns
    -------
    grid: (n, 2) list of 2D points
    '''
    x_grid = np.arange(*bounds[:,0], step = step)
    y_grid = np.arange(*bounds[:,1], step = step)
    grid   = np.dstack(np.meshgrid(x_grid, y_grid)).reshape((-1,2))
    return grid

def grid_linspace_2D(bounds, count):
    '''
    Return a count*count 2D grid

    Arguments
    ---------
    bounds: (2,2) list of [[minx, miny], [maxx, maxy]]
    count:  int, number of elements on a side
    
    Returns
    -------
    grid: (count**2, 2) list of 2D points
    '''
    x_grid = np.linspace(*bounds[:,0], count = count)
    y_grid = np.linspace(*bounds[:,1], count = count)
    grid   = np.dstack(np.meshgrid(x_grid, y_grid)).reshape((-1,2))
    return grid

def replace_references(data, reference_dict):
    # Replace references in place
    view = np.array(data).view().reshape((-1))
    for i, value in enumerate(view):
        if value in reference_dict:
            view[i] = reference_dict[value]
    return view

def multi_dict(pairs):
    '''
    Given a set of key value pairs, create a dictionary. 
    If a key occurs multiple times, stack the values into an array.

    Can be called like the regular dict(pairs) constructor

    Arguments
    ----------
    pairs: (n,2) array of key, value pairs

    Returns
    ----------
    result: dict, with all values stored (rather than last with regular dict)

    '''
    result = defaultdict(list)
    for k, v in pairs:
        result[k].append(v)
    return result

def tolist_dict(data):
    def tolist(item):
        if hasattr(item, 'tolist'):
            return item.tolist()
        else:
            return item
    result = {k:tolist(v) for k,v in data.items()}
    return result
    
def is_binary_file(file_obj):
    '''
    Returns True if file has non-ASCII characters (> 0x7F, or 127)
    Should work in both Python 2 and 3
    '''
    start  = file_obj.tell()
    fbytes = file_obj.read(1024)
    file_obj.seek(start)
    is_str = isinstance(fbytes, str)
    for fbyte in fbytes:
        if is_str: code = ord(fbyte)
        else:      code = fbyte
        if code > 127: return True
    return False

def decimal_to_digits(decimal, min_digits=None):
    digits = abs(int(np.log10(decimal)))
    if min_digits is not None:
        digits = np.clip(digits, min_digits, 20)
    return digits

def md5_object(obj):
    '''
    If an object is hashable, return the hex string of the MD5.
    '''
    hasher = hashlib.md5()
    hasher.update(obj)
    hashed = hasher.hexdigest()
    return hashed

def attach_to_log(log_level = logging.DEBUG, 
                  blacklist = ['TerminalIPythonApp','PYREADLINE']):
    '''
    Attach a stream handler to all loggers.
    '''
    try: 
        from colorlog import ColoredFormatter
        formatter = ColoredFormatter(
            ("%(log_color)s%(levelname)-8s%(reset)s " + 
             "%(filename)17s:%(lineno)-4s  %(blue)4s%(message)s"),
            datefmt = None,
            reset   = True,
            log_colors = {'DEBUG':    'cyan',
                          'INFO':     'green',
                          'WARNING':  'yellow',
                          'ERROR':    'red',
                          'CRITICAL': 'red' } )
    except ImportError: 
        formatter = logging.Formatter(
            "[%(asctime)s] %(levelname)-7s (%(filename)s:%(lineno)3s) %(message)s", 
            "%Y-%m-%d %H:%M:%S")
    handler_stream = logging.StreamHandler()
    handler_stream.setFormatter(formatter)
    handler_stream.setLevel(log_level)

    for logger in logging.Logger.manager.loggerDict.values():
        if (logger.__class__.__name__ != 'Logger' or  
            logger.name in blacklist):
            continue
        logger.addHandler(handler_stream)
        logger.setLevel(log_level)
    np.set_printoptions(precision=5, suppress=True)
    
def tracked_array(array, dtype=None):
    '''
    Properly subclass a numpy ndarray to track changes. 
    '''
    result = np.ascontiguousarray(array).view(TrackedArray)
    if dtype is None: 
        return result
    return result.astype(dtype)


class TrackedArray(np.ndarray):
    '''
    Track changes in a numpy ndarray.

    Methods
    ----------
    md5: returns hexadecimal string of md5 of array
    '''

    def __array_finalize__(self, obj):
        '''
        Sets a modified flag on every TrackedArray
        This flag will be set on every change, as well as during copies
        and certain types of slicing. 
        '''
        self._modified = True
        if isinstance(obj, type(self)):
            obj._modified = True
            
    def md5(self):
        '''
        Return an MD5 hash of the current array in hexadecimal string form. 
        
        This is quite fast; on a modern i7 desktop a (1000000,3) floating point 
        array was hashed reliably in .03 seconds. 
        
        This is only recomputed if a modified flag is set which may have false 
        positives (forcing an unnecessary recompute) but will not have false 
        negatives which would return an incorrect hash. 
        '''

        if self._modified or not hasattr(self, '_hashed'):
            self._hashed = md5_object(self)
        self._modified = False
        return self._hashed

    def __hash__(self):
        '''
        Hash is required to return an int, so we convert the hex string to int.
        '''
        return int(self.md5(), 16)
        
    def __setitem__(self, i, y):
        self._modified = True
        super(self.__class__, self).__setitem__(i, y)

    def __setslice__(self, i, j, y):
        self._modified = True
        super(self.__class__, self).__setslice__(i, j, y)

class Cache:
    '''
    Class to cache values until an id function changes.
    '''
    def __init__(self, id_function=None):
        if id_function is None:
            self._id_function = lambda: None
        else:
            self._id_function = id_function
        self.id_current   = None
        self._lock = 0
        self.cache = {}
        
    def get(self, key):
        '''
        Get a key from the cache.

        If the key is unavailable or the cache has been invalidated returns None.
        '''
        self.verify()
        if key in self.cache: 
            return self.cache[key]
        return None
        
    def verify(self):
        '''
        Verify that the cached values are still for the same value of id_function, 
        and delete all stored items if the value of id_function has changed. 
        '''
        id_new = self._id_function()
        if (self._lock == 0) and (id_new != self.id_current):
            if len(self.cache) > 0:
                log.warn('Clearing cache of %d items', len(self.cache))
                log.debug('Items cleared: %s', str(self.cache.keys()))
            self.clear()
            self.id_set()

    def clear(self, exclude=None):
        '''
        Remove all elements in the cache. 
        '''
        if exclude is None:
            self.cache = {}
        else:
            self.cache = {k:v for k,v in self.cache.items() if k in exclude}

    def update(self, items):
        '''
        Update the cache with a set of key, value pairs without checking id_function.
        '''
        self.cache.update(items)
        self.id_set()
       
    def id_set(self):
        self.id_current = self._id_function()

    def set(self, key, value):
        self.verify()
        self.cache[key] = value
        return value
        
    def __contains__(self, key):
        self.verify()
        return key in self.cache

    def __enter__(self):
        self._lock += 1
        
    def __exit__(self, *args):
        self._lock -= 1
        self.id_current = self._id_function()

def stack_lines(indices):
    return np.column_stack((indices[:-1],
                            indices[1:])).reshape((-1,2))

def append_faces(vertices_seq, faces_seq): 
    '''
    Given a sequence of zero- indexed faces and vertices,
    combine them into a single (n,3) list of faces and (m,3) vertices

    Arguments
    ---------
    vertices_seq: (n) sequence of (m,d) vertex arrays
    faces_seq     (n) sequence of (p,j) faces, zero indexed
                  and referencing their counterpoint vertices

    '''
    vertices_len = np.array([len(i) for i in vertices_seq])
    face_offset  = np.append(0, np.cumsum(vertices_len)[:-1])
    
    for offset, faces in zip(face_offset, faces_seq):
        faces += offset

    vertices = np.vstack(vertices_seq)
    faces    = np.vstack(faces_seq)

    return vertices, faces

def array_to_base64(array, dtype=None):
    '''
    Export a numpy array to a compact serializable dictionary.

    Arguments
    ---------
    array: numpy array
    dtype: optional, what dtype should array be encoded with.

    Returns
    ---------
    encoded: dict with keys: 
                 dtype: string of dtype
                 shape: int tuple of shape
                 base64: base64 encoded string of flat array
    '''
    array = np.asanyarray(array)
    shape = array.shape
    # ravel also forces contiguous
    flat  = np.ravel(array)
    if dtype is None:
        dtype = array.dtype
    encoded = {'dtype'  : np.dtype(dtype).str,
               'shape'  : shape,
               'base64' : base64.b64encode(flat.astype(dtype))}
    return encoded

def base64_to_array(encoded):
    '''
    Turn a dictionary with base64 encoded strings back into a numpy array.

    Arguments
    ----------
    encoded: dict with keys: 
                 dtype: string of dtype
                 shape: int tuple of shape
                 base64: base64 encoded string of flat array

    Returns
    ----------
    array: numpy array
    '''
    shape = encoded['shape']
    dtype = np.dtype(encoded['dtype'])
    array = np.fromstring(base64.b64decode(encoded['base64']), 
                          dtype).reshape(shape)

    return array

def is_instance_named(obj, name):
    '''
    Given an object, if it is a member of the class 'name',
    or a subclass of 'name', return True.

    Arguments
    ---------
    obj: instance of a class
    name: string

    Returns
    ---------
    bool, whether the object is a member of the named class
    '''

    # if obj is a member of the named class, return True
    if obj.__class__.__name__ == name:
        return True

    # if obj is a subclass of the name, return True
    return any(i.__name__ == name for i in obj.__class__.__bases__)

def submesh(mesh, 
            faces_sequence, 
            only_watertight = False, 
            append = False):
    '''
    Return a subset of the mesh.

    Arguments
    ----------
    mesh: Trimesh object
    faces_sequence: sequence of face indices from mesh
    only_watertight: only return submeshes which are watertight. 
    append: return a single mesh which has the faces specified appended.
            if this flag is set, only_watertight is ignored

    Returns
    ---------
    if append: Trimesh object
    else:      list of Trimesh objects
    '''

    # avoid nuking the cache on the original mesh
    original_faces    = mesh.faces.view(np.ndarray)
    original_vertices = mesh.vertices.view(np.ndarray)

    faces    = deque()
    vertices = deque()
    normals  = deque()
    visuals  = deque()

    # for reindexing faces
    mask = np.arange(len(original_vertices))    

    for faces_index in faces_sequence:        
        # sanitize indices in case they are coming in as a set or tuple
        faces_index   = np.array(list(faces_index))
        faces_current = original_faces[faces_index]
        unique = np.unique(faces_current.reshape(-1))
        
        # redefine face indices from zero
        mask[unique] = np.arange(len(unique))

        normals.append(mesh.face_normals[faces_index])
        faces.append(mask[faces_current])
        vertices.append(original_vertices[unique])
        visuals.extend(mesh.visual.subsets([faces_index]))

    if append:
        visuals = np.array(visuals)
        vertices, faces = append_faces(vertices, faces)
        appended = type(mesh)(vertices = vertices,
                              faces = faces,
                              face_normals = np.vstack(normals),
                              visual  = visuals[0].union(visuals[1:]),
                              process = False)
        return appended

    # we use type(mesh) rather than importing Trimesh from base
    # as this causes a circular import
    result = [type(mesh)(vertices     = v, 
                         faces        = f, 
                         face_normals = n,
                         visual       = c,
                         process = False) for v,f,n,c in zip(vertices, 
                                                             faces, 
                                                             normals,
                                                             visuals)]
    result = np.array(result)
    if only_watertight:
        watertight = np.array([i.fill_holes() and len(i.faces) > 4 for i in result])
        result     = result[watertight]
    return result



import numpy as np

from ..io.stl   import load_stl

from string     import Template
from tempfile   import NamedTemporaryFile
from subprocess import check_call

class MeshScript:
    def __init__(self, 
                 meshes, 
                 script):
        self.meshes  = meshes
        self.script  = script

    def __enter__(self):
        self.mesh_pre = [NamedTemporaryFile(suffix='.STL') for i in self.meshes]
        self.mesh_post  = NamedTemporaryFile(suffix='.STL')
        self.script_out = NamedTemporaryFile()

        # export the meshes to a temporary STL container
        for m, f in zip(self.meshes, self.mesh_pre):
            m.export(file_type='stl', file_obj=f.name)

        self.replacement = {'mesh_' + str(i) : m.name for i,m in enumerate(self.mesh_pre)}
        self.replacement['mesh_pre']  = str([i.name for i in self.mesh_pre])
        self.replacement['mesh_post'] = self.mesh_post.name
        self.replacement['script']    = self.script_out.name

        script_text = Template(self.script).substitute(self.replacement)
        self.script_out.write(script_text.encode('utf-8'))
        self.script_out.flush()

        return self

    def run(self, command):
        command_run = Template(command).substitute(self.replacement).split()
        # run the binary
        check_call(command_run)

        # bring the binaries result back as a Trimesh object
        self.mesh_post.seek(0)
        mesh_result = load_stl(self.mesh_post)
        return mesh_result
    
    def __exit__(self, *args, **kwargs):
        # close all the freaking temporary files
        self.mesh_post.close()
        self.script_out.close()
        for f in self.mesh_pre:
            f.close()



import numpy as np

from .constants import log, tol
from .geometry  import faces_to_edges
from .points    import  unitize, project_to_plane 

def mesh_plane_intersection(mesh, 
                            plane_normal,
                            plane_origin  = None,
                            return_planar = False):
    '''
    Return a cross section of the trimesh based on plane origin and normal. 
    Basically a bunch of plane-line intersection queries

    origin:        (3) array of plane origin
    normal:        (3) array for plane normal
    return_planar: bool, True returns:
                         (m,2,2) list of 2D line segments
                         False returns:
                         (m,2,3) list of 3D line segments
    '''
    if len(mesh.faces) == 0: 
        raise NameError("Cannot compute cross section of empty mesh.")
    if plane_origin is None: 
        plane_origin = [0,0,0]

    edges = np.sort(faces_to_edges(mesh.faces), axis=1)
    intersections, valid  = plane_line_intersection(plane_origin, 
                                                    plane_normal, 
                                                    mesh.vertices[edges.T],
                                                    line_segments = True)
    log.debug('mesh_cross_section found %i intersections', len(intersections))
    if return_planar:
        return project_to_plane(intersections.reshape((-1,3)),
                                plane_normal = plane_normal,
                                plane_origin = plane_origin).reshape((-1,2,2))
    return intersections.reshape(-1,2,3)

def plane_line_intersection(plane_origin, 
                            plane_normal, 
                            endpoints,                            
                            line_segments = True):
    '''
    Calculates plane-line intersections

    Arguments
    ---------
    plane_origin:  plane origin, (3) list
    plane_normal:  plane direction (3) list
    endpoints:     points defining lines to be intersected, (2,n,3)
    line_segments: if True, only returns intersections as valid if
                   vertices from endpoints are on different sides
                   of the plane.

    Returns
    ---------
    intersections: (m, 3) list of cartesian intersection points
    valid        : (n, 3) list of booleans indicating whether a valid
                   intersection occurred
    '''
    endpoints    = np.array(endpoints)
    line_dir     = unitize(endpoints[1] - endpoints[0])
    plane_normal = unitize(np.asanyarray(plane_normal).reshape(3))
    plane_origin = np.asanyarray(plane_origin).reshape(3)

    t = np.dot(plane_normal, np.transpose(plane_origin - endpoints[0]))
    b = np.dot(plane_normal, np.transpose(line_dir))
    
    # If the plane normal and line direction are perpendicular, it means
    # the vector is 'on plane', and there isn't a valid intersection.
    # We discard on-plane vectors by checking that the dot product is nonzero
    valid = np.abs(b) > tol.merge
    if line_segments:
        test = np.dot(plane_normal, np.transpose(plane_origin - endpoints[1]))
        different_sides = np.sign(t) != np.sign(test)
        nonzero = np.logical_or(np.abs(t) > tol.merge,
                                np.abs(test) > tol.merge)
        valid = np.logical_and(valid, different_sides)
        valid = np.logical_and(valid, nonzero)

    d  = np.divide(t[valid], b[valid])
    intersection  = endpoints[0][valid]
    intersection += np.reshape(d, (-1,1)) * line_dir[valid]

    return intersection, valid
